package com.sillycycle.bagleyd.life4d;

/*
 * @(#)Life4DFrame.java
 *
 * Copyright 2010 - 2021  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.sillycycle.bagleyd.util.ArgumentParser;
import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.ComponentUtil;
import com.sillycycle.bagleyd.util.Icon;
import com.sillycycle.bagleyd.util.IODialog;
import com.sillycycle.bagleyd.util.MenuScroller;
import com.sillycycle.bagleyd.util.OKDialog;
import com.sillycycle.bagleyd.util.Spinner;
import com.sillycycle.bagleyd.util.TrafficLight;

/**
 * The <code>Life4DFrame</code> class is to be embedded in a Web page
 * or run as an application.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/life.html
 */

public class Life4DFrame extends JFrame implements ActionListener {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final String[] COMMAND_STRINGS = new String[] {
		"Get Buffer", "Write Buffer", "Show Buffer",
		"Print", "Signature", "Sig List", "DB List", "Stamp",
		"Forms", "Pattern Search", "Glider Search", "Search Param %",
		"Stop", "Step (Yield)", "Run", "Clear",
		"Speed >", "Slow <", "Setup ~",
		"Description ?", "Features !", "References ^", "About",
		"Get File", "Write File", "Exit"};
	static final char[] COMMAND_CHARS = new char[] {
		'g', 'w', '*', 'p', 'n', '$', '@', 'm', 'f',
		't', 'l', '%', 's', 'y', 'r', 'c', '>', '<', '~',
		'?', '!', '^', 'a', 'G', 'W', 'x'};
	static final String[] TRAFFIC_STRINGS = new String[] {
		"Stop", "Step", "Run", "Clear"
	};
	static final char[] TRAFFIC_CHARS = new char[] {
		's', 'y', 'r', 'c'
	};
	static final Color[] TRAFFIC_COLORS = new Color[] {
		Color.red, Color.yellow, Color.green, Color.blue // For Jimmi
	};

	static ArgumentParser argumentParser = null;
	Icon mediumIcon = null;
	Icon largeIcon = null;
	int widthValue = 806;
	int heightValue = 596;
	static int xValue = 0;
	static int yValue = 0;
	Color bgColor = Color.gray.darker();
	boolean controlValue = true;
	int zoneValue = Life4DInterface.DEFAULT_ZONES / 2;
	int speedValue = Life4DInterface.DEFAULT_SPEED;
	boolean debug = false;
	Life4D game = null;
	JPanel topPanel = new JPanel();
	Spinner zoneChanger = null;
	Spinner speedChanger = null;
	//Slider zoneChanger = null;
	//Slider speedChanger = null;
	TrafficLight trafficLight = new TrafficLight(TRAFFIC_STRINGS,
		TRAFFIC_CHARS, TRAFFIC_COLORS);
	JTextField ruleTextField = new JTextField(15);
	JTextField genTextField = new JTextField(5);
	JTextField cellsTextField = new JTextField(5);
	transient Life4DRules lifeRules = null;
	transient Life4DForms lifeForms = null;
	OKDialog aDialog = null;
	StringBuilder stringBuilderSave;
	Pattern4DDialog patternDialog = null;
	Life4DDialog gameDialog = null;
	IODialog ioDialog = null;
	private static JMenu formsMenu = null;
	private static JMenu[] subFormsMenu = null;
	static int neighborIndex = Life4DInterface.DEFAULT_NEIGHBOR_INDEX;
	static int statesValue = 2;
	JMenuBar menuBar;
	JMenuItem detachMenuItem;
	JPanel mainPanel = new JPanel();
	JPanel secondaryPanel = new JPanel();
	JFrame frame;

	public String writeLifeString() {
		return this.stringBuilderSave.toString();
	}

	public void getLifeString(String s) {
		game.stringSave = s;
	}

	public void startGame() {
		game.start();
	}

	public void endGame() {
		game.setPause(true);
	}

	public void pauseGame() {
		game.setPause(!game.getPause());
	}

	public void clearGame() {
		game.clearField(true);
		game.start();
	}

	public void processKey(char key) {
		game.processKey(key);
	}

	public synchronized void setZone(int i) {
		this.zoneValue = i;
		if (zoneChanger != null)
			zoneChanger.setValue(i);
	}

	public synchronized void setDelay(int i) {
		this.speedValue = i;
		if (speedChanger != null)
			speedChanger.setValue(Life4DInterface.MAX_SPEED +
				Life4DInterface.MIN_SPEED - 1 - i);
	}

	public void setColumns(int i) {
		if (game == null) {
			gameDialog.setColumns(i);
		} else {
			game.setColumns(i);
			gameDialog.setColumns(game.getColumns());
		}
	}

	public static int getColumns() {
		return Life4DDialog.getColumns();
	}

	public void setExcludeCount(int i) {
		patternDialog.setExcludeCount(i);
		if (game != null)
			game.setExcludeCount(i);
	}

	public static int getExcludeCount() {
		return Pattern4DDialog.getExcludeCount();
	}

	public void setSearchCount(int i) {
		patternDialog.setSearchCount(i);
		if (game != null)
			game.setSearchCount(i);
	}

	public static int getSearchCount() {
		return Pattern4DDialog.getSearchCount();
	}

	public void setExcludePeriod(int i) {
		patternDialog.setExcludePeriod(i);
		if (game != null)
			game.setExcludePeriod(i);
	}

	public static int getExcludePeriod() {
		return Pattern4DDialog.getExcludePeriod();
	}

	public void setSearchLength(int i) {
		patternDialog.setSearchLength(i);
		if (game != null)
			game.setSearchLength(i);
	}

	public static int getSearchLength() {
		return Pattern4DDialog.getSearchLength();
	}

	public void setFillPercent(int i) {
		patternDialog.setFillPercent(i);
		if (game != null)
			game.setFillPercent(i);
	}

	public static int getFillPercent() {
		return Pattern4DDialog.getFillPercent();
	}

	public void setBlobPercent(int i) {
		patternDialog.setBlobPercent(i);
		if (game != null)
			game.setBlobPercent(i);
	}

	public static int getBlobPercent() {
		return Pattern4DDialog.getBlobPercent();
	}

	public void setStirs(int i) {
		patternDialog.setStirs(i);
		if (game != null)
			game.setStirs(i);
	}

	public static int getStirs() {
		return Pattern4DDialog.getStirs();
	}

	public void setNumberedSearch(int i) {
		patternDialog.setNumberedSearch(i);
		if (game != null)
			game.setNumberedSearch(i);
	}

	public static int getNumberedSearch() {
		return Pattern4DDialog.getNumberedSearch();
	}

	public void setStates(int i) {
		gameDialog.setStates(i);
		if (game != null)
			game.setStates(i);
	}

	public static int getStates() {
		return Life4DDialog.getStates();
	}

	public void setNeighborIndex(int i) {
		gameDialog.setNeighborIndex(i);
		if (game != null)
			game.setNeighborIndex(i);
	}

	public static int getNeighborIndex() {
		return Life4DDialog.getNeighborIndex();
	}

	public Integer getRuleIndex() {
		if (game != null)
			return game.getRuleIndex(neighborIndex);
		return Life4DRules.defaultRuleIndex(neighborIndex);
	}

	public void setRules(int oldNeighborIndex, int newNeighborIndex) {
		if (game != null) {
			Integer ruleIndex;

			game.setNeighbors(newNeighborIndex);
			//gameDialog.updateList();
			game.ruleChange();
			if (Life4DCanvas.polychoronChange(oldNeighborIndex, newNeighborIndex)
) {
				game.clearField(true);
				game.eraseFrame();
				game.resizeField();
			}
			ruleIndex = game.getRuleIndex(newNeighborIndex);
			subFormsMenu = updateForms(formsMenu, subFormsMenu,
				newNeighborIndex, ruleIndex);
		}
	}

	public void setRule(int neighborIndex, String rule) {
		int birth = Life4DRules.getBirth(rule);
		int survival = Life4DRules.getSurvival(rule);

		gameDialog.setMenu(neighborIndex, birth, survival);
		gameDialog.setBirth(birth);
		gameDialog.setSurvival(survival);
		String string = Integer.toString(Life4DRules.indexToNeighbor[neighborIndex]) +
			Life4DRules.parseRuleFull(rule);
		if (statesValue != 2)
			updateRule(Integer.toString(statesValue) + "N" + string);
		else
			updateRule(string);
	}

	public int getBirth() {
		return gameDialog.getBirth();
	}

	public int getSurvival() {
		return gameDialog.getSurvival();
	}

	public void setRandomOrient3D(boolean flag) {
		gameDialog.setRandomOrient3D(flag);
		if (game != null)
			game.setRandomOrient3D(flag);
	}

	public static boolean getRandomOrient3D() {
		return Life4DDialog.getRandomOrient3D();
	}

	public void setRandomOrient(boolean flag) {
		gameDialog.setRandomOrient(flag);
		if (game != null)
			game.setRandomOrient(flag);
	}

	public static boolean getRandomOrient() {
		return Life4DDialog.getRandomOrient();
	}

	public void setLissajous(boolean flag) {
		gameDialog.setLissajous(flag);
		if (game != null)
			game.setLissajous(flag);
	}

	public static boolean getLissajous() {
		return Life4DDialog.getLissajous();
	}

	public void setGrid(boolean flag) {
		gameDialog.setGrid(flag);
		if (game != null)
			game.setGrid(flag);
	}

	public static boolean getGrid() {
		return Life4DDialog.getGrid();
	}

	/*public void setInvert(boolean flag) {
		gameDialog.setInvert(flag);
		if (game != null)
			game.setInvert(flag);
	}

	public boolean getInvert() {
		return gameDialog.getInvert();
	}*/

	public void setWrap(boolean flag) {
		gameDialog.setWrap(flag);
		if (game != null)
			game.setWrap(flag);
	}

	public static boolean getWrap() {
		return Life4DDialog.getWrap();
	}

	public void setFindStable(boolean flag) {
		patternDialog.setFindStable(flag);
		if (game != null)
			game.setFindStable(flag);
	}

	public static boolean getFindStable() {
		return Pattern4DDialog.getFindStable();
	}

	public void setUnboundDetection(boolean flag) {
		patternDialog.setUnboundDetection(flag);
		if (game != null)
			game.setUnboundDetection(flag);
	}

	public static boolean getUnboundDetection() {
		return Pattern4DDialog.getUnboundDetection();
	}

	public void setEdgeDetection(boolean flag) {
		patternDialog.setEdgeDetection(flag);
		if (game != null)
			game.setEdgeDetection(flag);
	}

	public static boolean getEdgeDetection() {
		return Pattern4DDialog.getEdgeDetection();
	}

	public void setCellVisible(boolean flag) {
		gameDialog.setCellVisible(flag);
		if (game != null)
			game.setCellVisible(flag);
	}

	public static boolean getCellVisible() {
		return Life4DDialog.getCellVisible();
	}

	public void setMono(boolean flag) {
		gameDialog.setMono(flag);
		if (game != null) {
			game.setMono(flag);
		}
	}

	public static boolean getMono() {
		return Life4DDialog.getMono();
	}

	public void setRound(boolean flag) {
		gameDialog.setRound(flag);
		if (game != null) {
			game.setRound(flag);
		}
	}

	public static boolean getRound() {
		return Life4DDialog.getRound();
	}

	public void setWireframe(boolean flag) {
		gameDialog.setWireframe(flag);
		if (game != null) {
			game.setWireframe(flag);
		}
	}

	public static boolean getWireframe() {
		return Life4DDialog.getWireframe();
	}

	public String getOpt(String string) {
		return argumentParser.getOption(string);
	}

	public void getOpts() {
		int oldValue, newValue;
		int neighborsValue;
		int delayValue;
		int oldNeighborIndex;

		widthValue = ArgumentParser.parseOption(getOpt("windowWidth"),
			widthValue);
		heightValue = ArgumentParser.parseOption(getOpt("windowHeight"),
			heightValue);
		xValue = ArgumentParser.parseOption(getOpt("windowX"), xValue);
		yValue = ArgumentParser.parseOption(getOpt("windowY"), yValue);
		bgColor = ArgumentParser.parseOption(getOpt("bgColor"), bgColor);
		controlValue = ArgumentParser.parseOption(getOpt("control"),
			controlValue);
		delayValue = Life4DInterface.MAX_SPEED +
			Life4DInterface.MIN_SPEED - 1 - speedValue;
		delayValue = ArgumentParser.parseOption(getOpt("delay"),
			delayValue);
		speedValue = Life4DInterface.MAX_SPEED +
			Life4DInterface.MIN_SPEED - 1 - delayValue;
		oldValue = Life4DDialog.getColumns();
		newValue = ArgumentParser.parseOption(getOpt("columns"),
			oldValue);
		if (newValue < Life4DInterface.MIN_COLUMNS)
			newValue = Life4DInterface.MIN_COLUMNS;
		if (oldValue != newValue)
			gameDialog.setColumns(newValue);
		statesValue = ArgumentParser.parseOption(getOpt("states"),
			statesValue);
		gameDialog.setStates(statesValue);
		neighborsValue = ArgumentParser.parseOption(getOpt("neighbors"),
			Life4DRules.indexToNeighbor[Life4DDialog.getNeighborIndex()]);
		oldNeighborIndex = neighborIndex;
		neighborIndex = Life4DRules.neighborToIndex(neighborsValue);
		if (oldNeighborIndex != neighborIndex) {
			Integer ruleIndex = Life4DRules.defaultRuleIndex(neighborIndex);

			subFormsMenu = updateForms(formsMenu, subFormsMenu,
				neighborIndex, ruleIndex); // can set to 0 size menu
			gameDialog.setNeighborIndex(neighborIndex);
		}
		gameDialog.setRandomOrient3D(ArgumentParser.parseOption(getOpt("randomOrient3D"),
			Life4DDialog.getRandomOrient3D()));
		gameDialog.setRandomOrient(ArgumentParser.parseOption(getOpt("randomOrient"),
			Life4DDialog.getRandomOrient()));
		gameDialog.setLissajous(ArgumentParser.parseOption(getOpt("lissajous"),
			Life4DDialog.getLissajous()));
		gameDialog.setGrid(ArgumentParser.parseOption(getOpt("grid"),
			Life4DDialog.getGrid()));
		gameDialog.setWrap(ArgumentParser.parseOption(getOpt("wrap"),
			Life4DDialog.getWrap()));
		gameDialog.setCellVisible(ArgumentParser.parseOption(getOpt("visible"),
			Life4DDialog.getCellVisible()));
		gameDialog.setMono(ArgumentParser.parseOption(getOpt("mono"),
			Life4DDialog.getMono()));
		gameDialog.setRound(ArgumentParser.parseOption(getOpt("round"),
			Life4DDialog.getRound()));
		gameDialog.setWireframe(ArgumentParser.parseOption(getOpt("wireframe"),
			Life4DDialog.getWireframe()));
	}

	public void getOpts2() {
		int oldValue, newValue;
		String ruleString = getOpt("rule");

		if (ruleString != null) {
			String newRule;
			String neighborRule;
			int index = Life4DDialog.getNeighborIndex();

			newRule = Life4DRules.parse(
				Integer.toString(Life4DRules.indexToNeighbor[index]) +
				ruleString);
			neighborRule = Life4DRules.parseNeighborRule(newRule);
			setRule(index, neighborRule);
		}
		oldValue = Pattern4DDialog.getExcludeCount();
		newValue = ArgumentParser.parseOption(getOpt("excludeCount"), oldValue);
		if (newValue < 1 || newValue > 24)
			newValue = Life4DInterface.DEFAULT_EXCLUDE_COUNT;
		if (oldValue != newValue)
			setExcludeCount(newValue);
		oldValue = Pattern4DDialog.getSearchCount();
		newValue = ArgumentParser.parseOption(getOpt("searchCount"), oldValue);
		if (newValue < 8 || newValue > 256)
			newValue = Life4DInterface.DEFAULT_SEARCH_COUNT;
		if (oldValue != newValue)
			setSearchCount(newValue);
		oldValue = Pattern4DDialog.getExcludePeriod();
		newValue = ArgumentParser.parseOption(getOpt("excludePeriod"), oldValue);
		if (newValue < 1 || newValue > 240)
			newValue = Life4DInterface.DEFAULT_EXCLUDE_PERIOD;
		if (oldValue != newValue)
			setExcludePeriod(newValue);
		oldValue = Pattern4DDialog.getSearchLength();
		newValue = ArgumentParser.parseOption(getOpt("searchLength"), oldValue);
		if (newValue < 64 || newValue > 2 * Life4DInterface.DEFAULT_SEARCH_LENGTH)
			newValue = Life4DInterface.DEFAULT_SEARCH_LENGTH;
		if (oldValue != newValue)
			setSearchLength(newValue);
		oldValue = Pattern4DDialog.getFillPercent();
		newValue = ArgumentParser.parseOption(getOpt("fillPercent"), oldValue);
		if (newValue < 1 || newValue > 99)
			newValue = Life4DInterface.DEFAULT_FILL_PERCENT;
		if (oldValue != newValue)
			setFillPercent(newValue);
		oldValue = Pattern4DDialog.getBlobPercent();
		newValue = ArgumentParser.parseOption(getOpt("blobPercent"), oldValue);
		if (newValue < 1 || newValue > 99)
			newValue = Life4DInterface.DEFAULT_BLOB_PERCENT;
		if (oldValue != newValue)
			setBlobPercent(newValue);
		oldValue = Pattern4DDialog.getStirs();
		newValue = ArgumentParser.parseOption(getOpt("stirs"), oldValue);
		if (newValue < 0 || newValue > 10)
			newValue = Life4DInterface.DEFAULT_STIRS;
		if (oldValue != newValue)
			setStirs(newValue);
		setFindStable(ArgumentParser.parseOption(getOpt("findStable"),
			Pattern4DDialog.getFindStable()));
		setUnboundDetection(ArgumentParser.parseOption(getOpt("unboundDetection"),
			Pattern4DDialog.getUnboundDetection()));
		setEdgeDetection(ArgumentParser.parseOption(getOpt("edgeDetection"),
			Pattern4DDialog.getEdgeDetection()));
	}

	public void getOpts3() {
		int oldValue = Pattern4DDialog.getNumberedSearch();
		int newValue = ArgumentParser.parseOption(getOpt("numberedSearch"),
			oldValue);

		if (newValue < -1 || newValue >= Life4DSoup.symmetryStrings.length) {
			newValue = -1;
			if (oldValue != newValue)
				setNumberedSearch(newValue);
		}
		if (newValue != -1) {
			game.setSoupSearch(newValue);
			game.setPause(false);
		} else if (ArgumentParser.parseOption(getOpt("patternSearch"),
				false)) {
			game.setSoupSearch(true, false);
			game.setPause(false);
		} else if (ArgumentParser.parseOption(getOpt("gliderSearch"),
				false)) {
			game.setSoupSearch(false, true);
			game.setPause(false);
		 } else {
			String file0 = getOpt("file0");
			if (file0 == null || file0.equals(""))
				return;
			//String offset0 = getOpt("offset0"); // i.e. 0,0
			String file1 = getOpt("file1");
			//String offset1 = getOpt("offset1"); // i.e. 0,0
			 int shake = ArgumentParser.parseOption(getOpt("shake"), 0);
			game.setFilesSearch(file0, file1, shake);
			game.setSoupSearch(false, false);
			game.setPause(false);
		}
	}

	public void setupGame() {
		if (gameDialog != null) {
			gameDialog.setVisible(true);
		}
	}

	public void searchParamGame() {
		if (patternDialog != null) {
			patternDialog.setVisible(true);
		}
	}

	public void showBuffer() {
		if (ioDialog != null) {
			ioDialog.setVisible(true);
		}
	}

	public void showMessage(String msg) {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Life4D", msg, "OK",
				mediumIcon, false);
			aDialog.setBounds(this.getSize().width / 2,
				this.getSize().height / 2, 350, 120);
			aDialog.setVisible(true);
		}
	}

	public void descriptionHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Description",
"Implementation of 4d life." + NL +
"When running from web the \"write\" and \"read\" use cookies to output " + NL +
"data, else it uses \"" + Life4DCanvas.DATAFILE + "\" file." + NL +
"See https://www.sillycycle.com/lifeNotes.html for TODO list. " + NL +
"Probably lots of other things need to be done too, let me know." + NL,
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void featuresHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Features",
"Press mouse button on a on cell to erase (drag to do multiple)." + NL +
" " + NL +
"Press mouse button on a off cell to turn it on (drag to do multiple)." + NL +
" " + NL +
"Press \"g\" key to get (read) a saved life form to screen." + NL +
" " + NL +
"Press \"w\" key to save (write) a life form to screen." + NL +
" " + NL +
"Press \"G\" key to get (read) a saved life form to file." + NL +
" " + NL +
"Press \"W\" key to save (write) a life form to file." + NL +
" " + NL +
"Press \"P\" or \"p\" keys to print a life form." + NL +
" " + NL +
"Press \"C\" or \"c\" keys to clear." + NL +
" " + NL +
"Press \"T\" or \"t\" keys to do a pattern search." + NL +
"Use exclude count in setup to exclude a certain count or multiple." + NL +
"Use search count in setup largest cell count to search for." + NL +
"Use exclude period in setup to exclude certain periods." + NL +
"Use search in setup for length until period check." + NL +
" " + NL +
"Press \"L\" or \"l\" keys to do a glider search." + NL +
" " + NL +
"Press \"O\" or \"o\" keys to wrap." + NL +
" " + NL +
"Press \"J\" or \"j\" key to toggle lissajous viewpoint movement." + NL +
" " + NL +
"Press \"#\" or \".\" key to toggle grid." + NL +
" " + NL +
"Press \">\" or \".\" keys to speed up life." + NL +
" " + NL +
"Press \"<\" or \",\" keys to slow down life." + NL +
" " + NL +
"Press \"]\" or \"}\" keys to increase the column/row " +
"size." + NL +
" " + NL +
"Press \"[\" or \"{\" keys to decrease the column/row " +
"size." + NL +
" " + NL +
"Press \"Y\" or \"y\" keys to step." + NL +
" " + NL +
"Press \"R\", \"r\", or \"|\" key to toggle run." + NL +
" " + NL +
"Press \"S\" or \"s\" key to toggle stop run." + NL +
" " + NL +
"Press \"~\" key for setup." + NL +
" " + NL +
"Press \"%\" key for setup search parameters." + NL +
" " + NL +
"Press \"Esc\" key to hide and pause program." + NL +
" " + NL +
"Press \"Q\", \"q\", or \"CTRL-C\" keys to kill program if an application." + NL,
/* "The life4d may be resized.  Cells will reshape depending on" + NL +
"the room they have." + NL + */
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void referencesHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"References",
"xlock life4d source code" + NL +
"https://www.sillycycle.com/xlockmore.html" + NL +
" " + NL +
"Bays, Carter, \"A Note About the Discovery of Many New Rules for the" + NL +
"Game of 3D Life\", Complex Systems 16 (2006) 381-386" + NL +
" " + NL +
"Bays, Carter, \"Candidates for the Game of Life in Three Dimensions\"," + NL +
"1987, Complex Systems 1 373-400" + NL +
" " + NL +
"Bays, Carter, \"Patterns for Simple Cellular Automata in a Universe of" + NL +
"Dense-Packed Spheres\", 1987, Complex Systems 1 853-875" + NL +
" " + NL +
"Bays, Carter, \"A Note on the Discovery of a New Game of Three-" + NL +
"Dimensional Life\", 1988, Complex Systems 2 255-258" + NL +
" " + NL +
"Bays, Carter, \"A New Candidate Rule for the Game of Three-Dimensional" + NL +
"Life\", 1992, Complex Systems 6 433-441" + NL +
" " + NL +
"Bays, Carter, \"Further Notes on the Game of Three-Dimensional Life\"," + NL +
"1994, Complex Systems 8 67-73" + NL +
" " + NL +
"Bays, Carter, \"A Note About the Discovery of Many New Rules for the" + NL +
"Game of Three-Dimensional Life\", 2006, Complex Systems 16 381-386" + NL +
" " + NL +
"Dewdney, A.K., \"The Armchair Universe, Computer Recreations from the" + NL +
"Pages of Scientific American Magazine\", W.H. Freedman and Company," + NL +
"New York, 1988 (February 1987 p 16)" + NL +
" " + NL +
"Bays, Carter, \"The Game of Three Dimensional Life\", 86/11/20" + NL +
"with (latest?) update from 87/2/1, also http://www.cse.sc.edu/~bays/" + NL +
" " + NL +
"Meeker, Lee Earl, \"Four Dimensional Cellular Automata and the Game" + NL +
"of Life\" 1998 http://home.sc.rr.com/lmeeker/Lee/Home.html",
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public void aboutHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"About",
"Java Life4DFrame Version 8.3.0" + NL +
"Send bugs (reports or fixes) to the author:" +
"David Bagley <bagleyd AT verizon.net>" + NL +
"The latest version is at: " +
"https://www.sillycycle.com/life.html" + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public Icon getIcon(String string, int sx, int sy) {
		URL url = Life4DFrame.class.getResource("/" + string);

		if (url == null)
			return null;
		return new Icon(Toolkit.getDefaultToolkit().
			getImage(url), sx, sy);
	}

	@Override
	public void setSize(int width, int height) {
		if (game == null)
			return;
		if (width != widthValue || height != heightValue) {
			game.setBounds(0, 0, width, height);
			game.resizeField();
			widthValue = width;
			heightValue = height;
			game.drawField(true);
		}
		/*repaint();*/ // causes flashing
	}

	@Override
	public void paint(Graphics g) {
		setSize(secondaryPanel.getSize().width,
			secondaryPanel.getSize().height - topPanel.getSize().height);
		super.paint(g);
	}

	public String readTextArea() {
		return ioDialog.getText();
	}

	public void writeTextArea(String s) {
		ioDialog.setText(s);
	}

	public void shuffleDown() {
		ComponentUtil.findJFrame(this).toBack();
	}

	public void quit() {
		System.exit(0);
	}

	public void displayMessage(String string) {
		if (frame != null) {
			if (string == null || "".equals(string))
				frame.setTitle(Life4DInterface.TITLE);
			else
				frame.setTitle(Life4DInterface.TITLE +
					": " + string);
		}
	}

	@Override
	public void requestFocus() {
		if (game != null)
			game.requestFocus();
	}

	public void updateRule(String string) {
		if (ruleTextField == null)
			return;
		ruleTextField.setText(Life4DRules.parseRuleCondense(string));
	}

	public void updateGen(int gen) {
		if (genTextField == null)
			return;
		genTextField.setText(String.valueOf(gen));
	}

	public void updateCells(int cells) {
		if (cellsTextField == null)
			return;
		cellsTextField.setText(String.valueOf(cells));
	}

	public void initPopup() {
		if (game != null)
			game.initPopup();
	}

	public Image getImage(String fileName) {
		return Toolkit.getDefaultToolkit().getImage(fileName);
	}

	class PressActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			processKey(event.getActionCommand().charAt(0));
		}
	}

	class ZoneChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = zoneValue;

			zoneValue = zoneChanger.getValue();
			if (oldValue != zoneValue) {
				game.setZone(zoneValue);
			}
		}
	}

	class SpeedChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = speedValue;

			speedValue = speedChanger.getValue();
			if (oldValue != speedValue) {
				game.setDelay(Life4DInterface.MAX_SPEED +
					Life4DInterface.MIN_SPEED - 1 -
					speedValue);
			}
		}
	}

	public JMenu[] updateForms(JMenu topMenu,
			JMenu[] subMenu, int neighborhoodIndex,
			Integer ruleIndex) {
		JMenuItem menuItem;
		JMenu[] menu = subMenu;
		int categoryIndex;

		if (topMenu == null) {
			return new JMenu[]{};
		}
		if (menu == null) {
			menu = new
				JMenu[Life4DForms.Category.values().length + 1];
		}
		for (Life4DForms.Category c : Life4DForms.Category.values()) {
			if (menu.length == 0) // 0 size menu
				continue;
			categoryIndex = c.ordinal();
			if (menu[categoryIndex] != null)
				topMenu.remove(menu[categoryIndex]);
			/*if (subFormsMenu[categoryIndex] != null)
				formsMenu.remove(subFormsMenu[categoryIndex]);*/
			menu[categoryIndex] = new JMenu(title(c.toString()));
			// Says unused but needed for scrollable menus
			new MenuScroller(menu[categoryIndex], 32, 125, 0, 0);
			topMenu.add(menu[categoryIndex]);
			if (Life4DForms.data != null &&
					Life4DForms.data[neighborhoodIndex] != null &&
					Life4DForms.data[neighborhoodIndex].length > 0 &&
					ruleIndex != null) {
				int ri = ruleIndex;

				if (Life4DForms.data[neighborhoodIndex].length > ri &&
						Life4DForms.data[neighborhoodIndex][ri].length > 0 &&
						Life4DForms.data[neighborhoodIndex][ri][categoryIndex] != null) {
					for (LifeSet item : Life4DForms.data[neighborhoodIndex][ri][categoryIndex]) {
						menuItem = new JMenuItem(item.name);
						menu[categoryIndex].add(menuItem);
						menuItem.addActionListener(this);
						//System.out.println("neighborIndex: " + neighborIndex +
						//	", ruleIndex: " + ri +
						//	", categoryIndex: " + categoryIndex +
						//	", j: " + j);
					}
				}
			}
		}
		{
			int symStrings;
			categoryIndex = Life4DForms.Category.values().length;
			if (menu.length == 0) // 0 size menu
				return null;
			if (menu[categoryIndex] != null)
				topMenu.remove(menu[categoryIndex]);
			menu[categoryIndex] = new JMenu("Random");
			topMenu.add(menu[categoryIndex]);
			symStrings = Life4DSoup.symmetryStrings.length;
			for (int sym = 0; sym < symStrings; sym++) {
				menuItem = new JMenuItem(Life4DSoup.symmetryStrings[sym]);
				menu[categoryIndex].add(menuItem);
				menuItem.addActionListener(this);
			}
		}
		return menu;
	}

	public String title(String name) {
		if (name.length() > 1) {
			return name.substring(0, 1).toUpperCase() + name.substring(1).toLowerCase();
		}
		return name.toUpperCase();
	}

	@Override
	public void actionPerformed(ActionEvent event) {
		String command = event.getActionCommand();

		for (int opt = 0; opt < COMMAND_STRINGS.length; opt++) {
			if (command.equals(COMMAND_STRINGS[opt])) {
				processKey(COMMAND_CHARS[opt]);
				return;
			}
		}
		game.getFormFromMenu(command);
	}


	public Life4DFrame(String title) {
		JPanel controlPanel = new JPanel();
		JPanel textPanel = new JPanel();
		JPanel genPanel = new JPanel();
		JPanel canvasPanel = new JPanel();

		try {
			mediumIcon = getIcon(Life4DInterface.ICONS[Life4DInterface.Icons.X32.ordinal()], 32, 32);
			largeIcon = getIcon(Life4DInterface.ICONS[Life4DInterface.Icons.X48.ordinal()], 48, 48);
		} catch (Exception e) {
			largeIcon = null;
		}
		lifeRules = new Life4DRules(Life4DInterface.XML_FILE, null);
		lifeForms = new Life4DForms(Life4DInterface.XML_FILE, null);
		patternDialog = new Pattern4DDialog(this,
			ComponentUtil.findJFrame(this),
			Life4DInterface.TITLE + ": Search Param", "DONE");
		gameDialog = new Life4DDialog(this,
			ComponentUtil.findJFrame(this),
			Life4DInterface.TITLE + ": Setup", "DONE");
		ioDialog = new IODialog(ComponentUtil.findJFrame(this),
			Life4DInterface.TITLE + ": IO", "Life Form", "DONE");
		getOpts();
		topPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		controlPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 3, 3));
		textPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 3, 3));
		mainPanel.setLayout(new BorderLayout());
		mainPanel.setBackground(Color.white);
		if (controlValue) {
			menuBar = new JMenuBar();
			JMenu menu;
			JMenuItem menuItem;

			JPopupMenu.setDefaultLightWeightPopupEnabled(false);
			menu = new JMenu("File");
			menuItem = new JMenuItem("Get Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_G);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_W);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Show Buffer");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Print");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_P);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Signature");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_N);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Sig List");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("DB List");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Stamp");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_M);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Get File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Exit");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_X);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			formsMenu = new JMenu("Forms");
			if (subFormsMenu != null && subFormsMenu.length == 0)
				subFormsMenu = null; // recover from 0 size menu
			subFormsMenu = updateForms(formsMenu, subFormsMenu,
				getNeighborIndex(), getRuleIndex());
			menuItem.addActionListener(this);
			menuBar.add(Life4DFrame.formsMenu);
			menu = new JMenu("Search");
			menuItem = new JMenuItem("Pattern Search");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_T);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Glider Search");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_L);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Search Param %");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Play");
			menuItem = new JMenuItem("Stop");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_S);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Step (Yield)");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_Y);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Run");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_R);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Clear");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_C);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Speed >");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Slow <");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Setup ~");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Help");
			menuItem = new JMenuItem("Description ?");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Features !");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("References ^");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("About");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_A);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			//setJMenuBar(menuBar); // needs to be in a panel
			mainPanel.add(menuBar, BorderLayout.PAGE_START);
			controlPanel.add(trafficLight);
			genPanel.add(new JLabel("Rule:"));
			genPanel.add(ruleTextField);
			//ruleTextField.setEditable(false);
			genPanel.add(new JLabel("Generation:"));
			genPanel.add(genTextField);
			genTextField.setEditable(false);
			genPanel.add(new JLabel("Cells:"));
			genPanel.add(cellsTextField);
			cellsTextField.setEditable(false);
			zoneChanger = new Spinner("Zone", zoneValue,
				0, Life4DInterface.DEFAULT_ZONES - 1);
			speedChanger = new Spinner("Speed",
				speedValue, Life4DInterface.MIN_SPEED,
				Life4DInterface.MAX_SPEED);
			/*zoneChanger = new Slider("Zone", zoneValue,
				0, Life4DInterface.DEFAULT_ZONES - 1);
			zoneChanger.setWidth(Life4DInterface.DEFAULT_ZONES * 8);
			speedChanger = new Slider("Speed",
				speedValue, Life4DInterface.MIN_SPEED,
				Life4DInterface.MAX_SPEED);
			speedChanger.setWidth(120);*/
			textPanel.add(genPanel);
			controlPanel.add(zoneChanger);
			controlPanel.add(speedChanger);
			gameDialog.setNeighborIndex(neighborIndex);
			ruleTextField.setText(//Math.abs(Life4DRules.indexToNeighbor[neighborIndex]) +
				Life4DRules.indexToNeighbor[neighborIndex] +
				Life4DRules.defaultRuleStrings[neighborIndex]);
			genTextField.setText("0");
			cellsTextField.setText("0");
			topPanel.add(controlPanel);
			topPanel.add(textPanel);
			secondaryPanel.setLayout(new BorderLayout());
			secondaryPanel.add(topPanel, BorderLayout.PAGE_START);
		}
		game = new Life4D(this, speedValue,
			Life4DDialog.getColumns(), Life4DDialog.getStates(),
			Life4DRules.indexToNeighbor[Life4DDialog.getNeighborIndex()],
			Life4DDialog.getRandomOrient3D(),
			Life4DDialog.getRandomOrient(), Life4DDialog.getLissajous(),
			Life4DDialog.getGrid(), Life4DDialog.getWrap(),
			Life4DDialog.getCellVisible(), Life4DDialog.getMono(),
			Life4DDialog.getRound(), Life4DDialog.getWireframe(),
			bgColor);
		canvasPanel.add(game);
		secondaryPanel.add(canvasPanel, BorderLayout.CENTER);
		mainPanel.add(secondaryPanel, BorderLayout.CENTER);
		getContentPane().add(mainPanel, BorderLayout.CENTER);
		if (controlValue) {
			trafficLight.addActionListener(new PressActionListener());
			zoneChanger.addChangeListener(new ZoneChangeListener());
			speedChanger.addChangeListener(new SpeedChangeListener());
			ruleTextField.addActionListener((ActionEvent event) -> {
				String rule = ruleTextField.getText();
				String newRule = Life4DRules.parse(rule);
				String neighborRule = Life4DRules.parseNeighborRule(newRule);
				int newNeighborIndex = Life4DRules.parseNeighborFromRule(newRule);

				if (neighborIndex != newNeighborIndex) {
					setRules(neighborIndex, newNeighborIndex);
					game.drawField(true);
					neighborIndex = newNeighborIndex;
					gameDialog.setNeighborIndex(neighborIndex);
				}
				setRule(newNeighborIndex, neighborRule);
			});
		}
		getOpts2();
		game.setBounds(0, 0, widthValue, heightValue);
		//show();
		setTitle(title);
		game.requestFocus();
		game.setPause(true);
		/* game.setSize(getSize().width,
			getSize().height - topPanel.getSize().height); */
		//game.setSize(widthValue, heightValue);
		getOpts3();
	}

	public static void main(String[] args) {
		String laf = UIManager.getCrossPlatformLookAndFeelClassName();

		try {
			UIManager.setLookAndFeel(laf);
		} catch (ClassNotFoundException | InstantiationException | IllegalAccessException | UnsupportedLookAndFeelException e) {
			e.printStackTrace();
			System.exit(1);
		}
		List<Image> iconList = new ArrayList<>();
		for (String icon : Life4DInterface.ICONS) {
			URL url = Life4DFrame.class.getResource("/" + icon);
			if (url != null)
				iconList.add(Toolkit.getDefaultToolkit().getImage(url));
		}

		argumentParser = new ArgumentParser(args);
		if (argumentParser.hasOption("?")) {
			System.out.println("usage:" + NL +
"\t[-windowWidth={int}] [-windowHeight={int}]" + NL +
"\t[-windowX={int}] [-windowY={int}]" + NL +
"\t[-bgColor={int}] [-control={bool}]" + NL +
"\t[-columns={int}] [-states={int}]" + NL +
"\t[-neighbors={int}] [-rule={string}]" + NL +
"\t[-excludeCount={int}] [-searchCount={int}]" + NL +
"\t[-excludePeriod={int}] [-searchLength={int}]" + NL +
"\t[-randomOrient3D={bool}] [-randomOrient={bool}]" + NL +
"\t[-lissajous={bool}] [-grid={bool}] [-wrap={bool}]" + NL +
"\t[-visible={bool}] [-mono={bool}] [-round={bool}]" + NL +
"\t[-wireframe={bool}] [-findStable={bool}]" + NL +
"\t[-unboundDetection={bool}] [-edgeDetection={bool}]" + NL +
"\t[-numberedSearch={int}] [-patternSearch={bool}]" + NL +
"\t[-gliderSearch={bool}] [-delay={int}]");
			return;
		}
		Life4DFrame frame = new Life4DFrame(Life4DInterface.TITLE);
		frame.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});
		if (!iconList.isEmpty()) {
			frame.setIconImages(iconList);
		}
		frame.pack();
		frame.game.requestFocusInWindow();
		frame.setLocation(Life4DFrame.xValue, Life4DFrame.yValue);
		frame.setVisible(true);
	}
}
