package com.sillycycle.bagleyd.util;

import java.awt.Dimension;

public class Dimension4D extends Dimension {
	private static final long serialVersionUID = 42L;
	public int depth, zone;


	/**
	 * Creates an instance of <code>Dimension</code> whose width
	 * and height are the same as for the specified dimension.
	 *
	 * @param width
	 * @param height
	 * @param depth
	 * @param zone
	 *
	 * the specified dimension for the
	 *               <code>width</code>,
	 *               <code>height</code>, and
	 *               <code>depth</code>, and
	 *               <code>zone</code> values
	 */
	public Dimension4D(int width, int height, int depth, int zone) {
		super.width = width;
		super.height = height;
		this.depth = depth;
		this.zone = zone;
	}

	/**
	 * Creates an instance of <code>Dimension</code> with a width
	 * of zero, a height of zero, a depth of zero, and a zone of zero.
	 */
	public Dimension4D() {
		super.width = 0;
		super.height = 0;
		this.depth = 0;
		this.zone = 0;
	}

	/**
	 * Checks whether two dimension objects have equal values.
	 *
	 * @return boolean
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj instanceof Dimension4D) {
			Dimension4D d = (Dimension4D) obj;
			return (width == d.width) && (height == d.height) &&
				(depth == d.depth) && (zone == d.zone);
		}
		return false;
	}

	/**
	 * Returns the hash code for this <code>Dimension4D</code>.
	 *
	 * @return a hash code for this <code>Dimension4D</code>
	 */
	@Override
	public int hashCode() {
		int sum = width + height + depth + zone;
		return sum * (sum + 3)/4 + width;
	}
}
