package com.sillycycle.bagleyd.util;

/*
 * @(#)Graphics3D.java
 *
 * Copyright 2009 - 2017  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.util.Random;

/*
 * Basic graphics algorithm for this was incorporated into
 * life3d.c in xlock in 1994.  Thought of using Java3D but it did
 * not look like the way to go... some examples did not seem to work
 * and many broken links.  I also want it fast and want to display
 * relatively simple 3D graphics. April 2009
 * Copyright 1993 Anthony Wesley awesley AT canb.auug.org.au found at
 * life.anu.edu.au /pub/complex_systems/alife/3DLIFE.ZIP
 */

/**
 * The <code>Cell</code> class handles the graphics info for
 * the screen.
 *
 * @author      David A. Bagley
 * @author      bagleyd AT verizon.net
 * @author      https://www.sillycycle.com/life.html
 */

public class Graphics3D {
	static final double EyeToScreen = 72.0;
	static final double HalfScreenD = 14.0;
	static final double PI180 = 3.1415926535 / 180.0;
	static final double LEN = 0.45;
	static final double LEN_2 = 0.225;
	static final double LEN2 = 0.9;
	static final double LEN4 = 1.8;
	static final int MAX_COLORS = 3; /* true for all Cb, Dd, Td */
	static final int MAX_VERTEX = 24; /* 8 if just Cb, 14 if just Dd */

	public static final int XN = 0;
	public static final int XP = 1;
	public static final int YN = 2;
	public static final int YP = 3;
	public static final int ZN = 4;
	public static final int ZP = 5;
	public static final int XNYNZN = 6; /* Truncated Octahedron */
	public static final int XPYNZN = 7;
	public static final int XNYPZN = 8;
	public static final int XPYPZN = 9;
	public static final int XNYNZP = 10;
	public static final int XPYNZP = 11;
	public static final int XNYPZP = 12;
	public static final int XPYPZP = 13;
	public static final int XNYN = 0; /* Rhombic Dodecahedron */
	public static final int XPYN = 1;
	public static final int XNYP = 2;
	public static final int XPYP = 3;
	public static final int YNZN = 4;
	public static final int YPZN = 5;
	public static final int YNZP = 6;
	public static final int YPZP = 7;
	public static final int ZNXN = 8;
	public static final int ZPXN = 9;
	public static final int ZNXP = 10;
	public static final int ZPXP = 11;

	Point[] pts = new Point[MAX_VERTEX]; /* screen coords for point */
	Point[] faceList = new Point[6];
	Graphics g;
	RealPoint v, o;
	int width, height, depth;
	double A, B, C, F;
	double metaAlt, metaAzm, metaDist;

	public Graphics3D(Graphics g, int ncols, int nrows, int nstacks,
			int width, int height, int depth) {
		Random generator = new Random(System.nanoTime());
		this.metaDist = generator.nextInt(360);
		this.metaAlt = generator.nextInt(360);
		this.metaAzm = generator.nextInt(360);
		this.o = new RealPoint(ncols / 2, nrows / 2, nstacks / 2);
		this.width = width;
		this.height = height;
		this.depth = depth;
		this.g = g;
		lissajous();
	}

	public Graphics getGraphics() {
		return g;
	}

	void newViewpoint(RealPoint vertex) {
		double k, l, d1, d2;

		k = vertex.x * vertex.x + vertex.y * vertex.y;
		l = Math.sqrt(k + vertex.z * vertex.z);
		k = Math.sqrt(k);
		d1 = (EyeToScreen / HalfScreenD);
		d2 = EyeToScreen / (HalfScreenD * this.height / this.width);
		this.A = d1 * l * (this.width / 2) / k;
		this.B = l * l;
		this.C = d2 * (this.height / 2) / k;
		this.F = k * k;
	}

	public void lissajous() {
		double alt, azm, dist;

		alt = 30.0 * Math.sin(this.metaAlt * PI180) + 45.0;
		this.metaAlt += 1.123;
		if (this.metaAlt >= 360.0)
			this.metaAlt -= 360.0;
		if (this.metaAlt < 0.0)
			this.metaAlt += 360.0;
		azm = 30.0 * Math.sin(this.metaAzm * PI180) + 45.0;
		this.metaAzm += 0.987;
		if (this.metaAzm >= 360.0)
			this.metaAzm -= 360.0;
		if (this.metaAzm < 0.0)
			this.metaAzm += 360.0;
		dist = 10.0 * Math.sin(this.metaDist * PI180) + 50.0;
		this.metaDist += 1.0;
		if (this.metaDist >= 360.0)
			this.metaDist -= 360.0;
		if (this.metaDist < 0.0)
			this.metaDist += 360.0;
		/*
		if (alt >= 90.0)
			alt = 90.0;
		else if (alt < -90.0)
			alt = -90.0;
		*/
		//this.azm = azm;
		/*System.out.println("dist " + dist + ", alt " + alt + ", azm " + azm);*/
		this.v = new RealPoint(
			Math.sin(azm * PI180) * Math.cos(alt * PI180) * dist,
			Math.cos(azm * PI180) * Math.cos(alt * PI180) * dist,
			Math.sin(alt * PI180) * dist);
		newViewpoint(this.v);
	}

	Point newPoint(RealPoint p) {
		double p1 = p.x * this.v.x + p.y * this.v.y;
		double E = this.B - p1 - p.z * this.v.z;

		return new Point((int) (this.width / 2 - this.A *
			(this.v.x * p.y - this.v.y * p.x) / E),
			(int) (this.height / 2 - this.C *
			(p.z * this.F - this.v.z * p1) / E));
	}

	/* float because double takes up too much memory and do not need */
	/* sqrt is not really needed for sort */
	public float getDistance(int i, int j, int k) {
		double dx = this.v.x - (i - o.x);
		double dy = this.v.y - (j - o.y);
		double dz = this.v.z - (k - o.z);
//System.out.println("distance_v:" + v.x + " " + v.y + " " + v.z);
//System.out.println("distance_o:" + (i - o.x) + " " + (i - o.y) + " " + (i - o.z));
//System.out.println("distance_d:" + dx + " " + dy + " " + dz);
		//return (float) Math.sqrt(dx * dx + dy * dy + dz * dz);
		return (float) (dx * dx + dy * dy + dz * dz);
	}

	/* Pure colors can get brighter */
	static Color brighter(Color color) {
		int r = color.getRed();
		int g = color.getGreen();
		int b = color.getBlue();
		final double FACTOR = 0.8;

		/* From 2D group:
		 * 1. black.brighter() should return grey
		 * 2. applying brighter to blue will always return blue, brighter
		 * 3. non pure color (non zero rgb) will eventually return white
		 */

		int i = (int) ((1 - FACTOR) * 255);

		if (r >= 0 && r < i)
			r = i;
		if (g >= 0 && g < i)
			g = i;
		if (b >= 0 && b < i)
			b = i;
		return new Color(Math.min((int) (r / FACTOR), 255),
			Math.min((int) (g / FACTOR), 255),
			Math.min((int) (b / FACTOR), 255));
	}

	static Color darker(Color color) {
		return color.darker();
	}

	boolean inTri(Point pt, Point[] triPts,
			int p0, int p1, int p2) {
		faceList[0] = triPts[p0];
		faceList[1] = triPts[p1];
		faceList[2] = triPts[p2];
		return pointInPolygon(pt, faceList, 3);
	}

	void drawTri(boolean wireframe, Color cid, Color bid, Point[] triPts,
			int p0, int p1, int p2) {
		int[] faceListX = new int[3];
		int[] faceListY = new int[3];
		faceListX[0] = triPts[p0].x;
		faceListY[0] = triPts[p0].y;
		faceListX[1] = triPts[p1].x;
		faceListY[1] = triPts[p1].y;
		faceListX[2] = triPts[p2].x;
		faceListY[2] = triPts[p2].y;
/*for (int i = 0; i < 3; i++)
System.out.println(i + ": " + faceListX[i] + " " + faceListY[i]);*/
		if (cid != null) {
			this.g.setColor(cid);
			if (wireframe) {
				this.g.drawPolygon(faceListX, faceListY, 3);
				return;
			}
			this.g.fillPolygon(faceListX, faceListY, 3);
		}
		if (bid != null || cid == null) {
			if (bid != null)
				this.g.setColor(bid);
			this.g.drawPolygon(faceListX, faceListY, 3);
		}
	}

	boolean inQuad(Point pt, Point[] quadPts,
			int p0, int p1, int p2, int p3) {
		faceList[0] = quadPts[p0];
		faceList[1] = quadPts[p1];
		faceList[2] = quadPts[p2];
		faceList[3] = quadPts[p3];
		return pointInPolygon(pt, faceList, 4);
	}

	void drawMouth(boolean wireframe, Color cid, Point[] quadPts,
			int[] p) {
		int size = p.length;
		int[] faceListX = new int[size];
		int[] faceListY = new int[size];
		int avX = 0, avY = 0;
		for (int i = 0; i < size; i++) {
			avX += quadPts[p[i]].x;
			avY += quadPts[p[i]].y;
		}
		avX /= size;
		avY /= size;
		for (int i = 0; i < size; i++) {
			faceListX[i] = (avX + quadPts[p[i]].x) / 2;
			faceListY[i] = (avY + quadPts[p[i]].y) / 2;
		}
/*for (int i = 0; i < size; i++)
System.out.println(i + ": " + faceListX[i] + " " +  faceListY[i]);*/
		if (cid != null) {
			this.g.setColor(cid);
			if (wireframe) {
				this.g.drawPolygon(faceListX, faceListY, size);
				return;
			}
			this.g.fillPolygon(faceListX, faceListY, size);
		}
	}

	void drawQuad(boolean wireframe, Color cid, Color bid, Point[] quadPts,
			int p0, int p1, int p2, int p3) {
		int[] faceListX = new int[4];
		int[] faceListY = new int[4];
		faceListX[0] = quadPts[p0].x;
		faceListY[0] = quadPts[p0].y;
		faceListX[1] = quadPts[p1].x;
		faceListY[1] = quadPts[p1].y;
		faceListX[2] = quadPts[p2].x;
		faceListY[2] = quadPts[p2].y;
		faceListX[3] = quadPts[p3].x;
		faceListY[3] = quadPts[p3].y;
/*for (int i = 0; i < 4; i++)
System.out.println(i + ": " + faceListX[i] + " " + faceListY[i]);*/
		if (cid != null) {
			this.g.setColor(cid);
			if (wireframe) {
				this.g.drawPolygon(faceListX, faceListY, 4);
				return;
			}
			this.g.fillPolygon(faceListX, faceListY, 4);
		}
		if (bid != null || cid == null) {
			if (bid != null)
				this.g.setColor(bid);
			this.g.drawPolygon(faceListX, faceListY, 4);
		}
	}

	void drawQuadMouth(boolean wireframe, Color cid, Color bid, Point[] quadPts,
			int p0, int p1, int p2, int p3, boolean mouth) {
		int[] faceListX = new int[4];
		int[] faceListY = new int[4];
		faceListX[0] = quadPts[p0].x;
		faceListY[0] = quadPts[p0].y;
		faceListX[1] = quadPts[p1].x;
		faceListY[1] = quadPts[p1].y;
		faceListX[2] = quadPts[p2].x;
		faceListY[2] = quadPts[p2].y;
		faceListX[3] = quadPts[p3].x;
		faceListY[3] = quadPts[p3].y;
/*for (int i = 0; i < 4; i++)
System.out.println(i + ": " + faceListX[i] + " " +  faceListY[i]);*/
		if (cid != null) {
			this.g.setColor(cid);
			if (wireframe) {
				this.g.drawPolygon(faceListX, faceListY, 4);
			} else {
				this.g.fillPolygon(faceListX, faceListY, 4);
			}
		}
		if (!wireframe && (bid != null || cid == null)) {
			if (bid != null)
				this.g.setColor(bid);
			this.g.drawPolygon(faceListX, faceListY, 4);
		}
		if (mouth) {
			int[] p = new int[4];
			p[0] = p0;
			p[1] = p1;
			p[2] = p2;
			p[3] = p3;
			drawMouth(wireframe, bid, quadPts, p);
		}
	}

	boolean inHex(Point pt, Point[] hexPts,
			int p0, int p1, int p2, int p3, int p4, int p5) {
		faceList[0] = hexPts[p0];
		faceList[1] = hexPts[p1];
		faceList[2] = hexPts[p2];
		faceList[3] = hexPts[p3];
		faceList[4] = hexPts[p4];
		faceList[5] = hexPts[p5];
		return pointInPolygon(pt, faceList, 6);
	}

	void drawHex(boolean wireframe, Color cid, Color bid, Point[] hexPts,
			int p0, int p1, int p2, int p3, int p4, int p5) {
		int[] faceListX = new int[6];
		int[] faceListY = new int[6];

		faceListX[0] = hexPts[p0].x;
		faceListY[0] = hexPts[p0].y;
		faceListX[1] = hexPts[p1].x;
		faceListY[1] = hexPts[p1].y;
		faceListX[2] = hexPts[p2].x;
		faceListY[2] = hexPts[p2].y;
		faceListX[3] = hexPts[p3].x;
		faceListY[3] = hexPts[p3].y;
		faceListX[4] = hexPts[p4].x;
		faceListY[4] = hexPts[p4].y;
		faceListX[5] = hexPts[p5].x;
		faceListY[5] = hexPts[p5].y;
/*for (int i = 0; i < 6; i++)
System.out.println(i + ": " + faceListX[i] + " " + faceListY[i]);*/
		if (cid != null) {
			this.g.setColor(cid);
			if (wireframe) {
				this.g.drawPolygon(faceListX, faceListY, 6);
				return;
			}
			this.g.fillPolygon(faceListX, faceListY, 6);
		}
		if (bid != null || cid == null) {
			if (bid != null)
				this.g.setColor(bid);
			this.g.drawPolygon(faceListX, faceListY, 6);
		}
	}

	void drawHexMouth(boolean wireframe, Color cid, Color bid, Point[] hexPts,
			int p0, int p1, int p2, int p3, int p4, int p5, boolean mouth) {
		int[] faceListX = new int[6];
		int[] faceListY = new int[6];
		faceListX[0] = hexPts[p0].x;
		faceListY[0] = hexPts[p0].y;
		faceListX[1] = hexPts[p1].x;
		faceListY[1] = hexPts[p1].y;
		faceListX[2] = hexPts[p2].x;
		faceListY[2] = hexPts[p2].y;
		faceListX[3] = hexPts[p3].x;
		faceListY[3] = hexPts[p3].y;
		faceListX[4] = hexPts[p4].x;
		faceListY[4] = hexPts[p4].y;
		faceListX[5] = hexPts[p5].x;
		faceListY[5] = hexPts[p5].y;
/*for (int i = 0; i < 6; i++)
System.out.println(i + ": " + faceListX[i] + " " +  faceListY[i]);*/
		if (cid != null) {
			this.g.setColor(cid);
			if (wireframe) {
				this.g.drawPolygon(faceListX, faceListY, 6);
			} else {
				this.g.fillPolygon(faceListX, faceListY, 6);
			}
		}
		if (!wireframe && (bid != null || cid == null)) {
			if (bid != null)
				this.g.setColor(bid);
			this.g.drawPolygon(faceListX, faceListY, 6);
		}
		if (mouth) {
			int[] p = new int[6];
			p[0] = p0;
			p[1] = p1;
			p[2] = p2;
			p[3] = p3;
			p[4] = p4;
			p[5] = p5;
			drawMouth(wireframe, bid, hexPts, p);
		}
	}

	/* Sphere for cube */
	void drawSphere(boolean w, Color[] cid, Color bid, Point3D cell) {
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y, cell.z - this.o.z);
		RealPoint d = new RealPoint();
		int absSize = 0, avgX = 0, avgY= 0;
		int i = 0, out = 0;

		for (d.z = b.z - LEN; d.z <= b.z + LEN2; d.z += LEN2)
		for (d.y = b.y - LEN; d.y <= b.y + LEN2; d.y += LEN2)
		for (d.x = b.x - LEN; d.x <= b.x + LEN2; d.x += LEN2) {
				pts[i] = this.newPoint(d);
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
				if (i > 0) {
					int temp;

					if (pts[i].x >= pts[0].x) {
						temp = pts[i].x - pts[0].x;
					} else {
						temp = pts[0].x - pts[i].x;
					}
					if (temp > absSize) {
						absSize = temp;
					}
					if (pts[i].y >= pts[0].y) {
						temp = pts[i].y - pts[0].y;
					} else {
						temp = pts[0].y - pts[i].y;
					}
					if (temp > absSize) {
						absSize = temp;
					}
				}
				avgX += pts[i].x;
				avgY += pts[i].y;
				++i;
			}
		absSize = absSize * 11 / 10;
		if (out == 8)
			return;
		avgX /= 8;
		avgY /= 8;
		if (w) {
			this.g.setColor(cid[1]);
		} else {
			this.g.setColor(bid);
		}
		this.g.drawOval(avgX - absSize / 2, avgY - absSize / 2,
			absSize, absSize);
		this.g.drawOval(avgX - absSize / 2, avgY - absSize / 2 - 1,
			absSize, absSize);
		this.g.drawOval(avgX - absSize / 2 - 1, avgY - absSize / 2,
			absSize, absSize);
		if (w || (cid[0] == null && cid[1] == null && cid[2] == null)) {
			return;
		}
		this.g.setColor(cid[0]);
		this.g.fillOval(avgX - absSize / 2, avgY - absSize / 2,
			absSize, absSize);
		absSize /= 3;
		avgX -= absSize / 2;
		avgY -= absSize / 2;
		this.g.setColor(cid[1]);
		this.g.fillOval(avgX - absSize / 2, avgY - absSize / 2,
			absSize, absSize);
		absSize /= 3;
		avgX -= absSize / 2;
		avgY -= absSize / 2;
		this.g.setColor(cid[2]);
		this.g.fillOval(avgX - absSize / 2, avgY - absSize / 2,
			absSize, absSize);
	}

	void drawSphere2(boolean w, Color[] cid, Color bid, Point3D cell) {
		int absSize = 0, avgX = 0, avgY= 0;
		int i, out = 0;
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y, (cell.z - this.o.z) / 2.0);

		if ((cell.z & 1) == 1) {
			b.x += 0.5;
			b.y += 0.5;
		}
		pts[0] = this.newPoint(new RealPoint(b, LEN_2, 0., LEN));
		pts[1] = this.newPoint(new RealPoint(b, 0., LEN_2, LEN));
		pts[2] = this.newPoint(new RealPoint(b, -LEN_2, 0., LEN));
		pts[3] = this.newPoint(new RealPoint(b, 0., -LEN_2, LEN));

		pts[4] = this.newPoint(new RealPoint(b, LEN, 0., LEN_2));
		pts[15] = this.newPoint(new RealPoint(b, LEN, -LEN_2, 0.));
		pts[16] = this.newPoint(new RealPoint(b, LEN, 0., -LEN_2));
		pts[8] = this.newPoint(new RealPoint(b, LEN, LEN_2, 0.));

		pts[5] = this.newPoint(new RealPoint(b, 0., LEN, LEN_2));
		pts[9] = this.newPoint(new RealPoint(b, LEN_2, LEN, 0.));
		pts[17] = this.newPoint(new RealPoint(b, 0., LEN, -LEN_2));
		pts[10] = this.newPoint(new RealPoint(b, -LEN_2, LEN, 0.));

		pts[6] = this.newPoint(new RealPoint(b, -LEN, 0., LEN_2));
		pts[11] = this.newPoint(new RealPoint(b, -LEN, LEN_2, 0.));
		pts[18] = this.newPoint(new RealPoint(b, -LEN, 0., -LEN_2));
		pts[12] = this.newPoint(new RealPoint(b, -LEN, -LEN_2, 0.));

		pts[7] = this.newPoint(new RealPoint(b, 0., -LEN, LEN_2));
		pts[13] = this.newPoint(new RealPoint(b, -LEN_2, -LEN, 0.));
		pts[19] = this.newPoint(new RealPoint(b, 0., -LEN, -LEN_2));
		pts[14] = this.newPoint(new RealPoint(b, LEN_2, -LEN, 0.));

		pts[20] = this.newPoint(new RealPoint(b, LEN_2, 0., -LEN));
		pts[21] = this.newPoint(new RealPoint(b, 0., LEN_2, -LEN));
		pts[22] = this.newPoint(new RealPoint(b, -LEN_2, 0., -LEN));
		pts[23] = this.newPoint(new RealPoint(b, 0., -LEN_2, -LEN));
		for (i = 0; i < 24; i++) {
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
				if (i > 0) {
					int temp;

					if (pts[i].x >= pts[0].x) {
						temp = pts[i].x - pts[0].x;
					} else {
						temp = pts[0].x - pts[i].x;
					}
					if (temp > absSize) {
						absSize = temp;
					}
					if (pts[i].y >= pts[0].y) {
						temp = pts[i].y - pts[0].y;
					} else {
						temp = pts[0].y - pts[i].y;
					}
					if (temp > absSize) {
						absSize = temp;
					}
				}
				avgX += pts[i].x;
				avgY += pts[i].y;
		}
		if (out == 24)
			return;
		absSize = absSize * 11 / 10;
		avgX /= 24;
		avgY /= 24;
		if (w) {
			this.g.setColor(cid[1]);
		} else {
			this.g.setColor(bid);
		}
		this.g.drawOval(avgX - absSize / 2, avgY - absSize / 2,
			absSize, absSize);
		this.g.drawOval(avgX - absSize / 2, avgY - absSize / 2 - 1,
			absSize, absSize);
		this.g.drawOval(avgX - absSize / 2 - 1, avgY - absSize / 2,
			absSize, absSize);
		if (w || (cid[0] == null && cid[1] == null && cid[2] == null)) {
			return;
		}
		this.g.setColor(cid[0]);
		this.g.fillOval(avgX - absSize / 2, avgY - absSize / 2,
			absSize, absSize);
		absSize /= 3;
		avgX -= absSize / 2;
		avgY -= absSize / 2;
		this.g.setColor(cid[1]);
		this.g.fillOval(avgX - absSize / 2, avgY - absSize / 2,
			absSize, absSize);
		absSize /= 3;
		avgX -= absSize / 2;
		avgY -= absSize / 2;
		this.g.setColor(cid[2]);
		this.g.fillOval(avgX - absSize / 2, avgY - absSize / 2,
			absSize, absSize);
	}

	boolean inCube(Point pt, Point3D cell) {
		int i = 0, out = 0;
		RealPoint d = new RealPoint();
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y, cell.z - this.o.z);

		for (d.z = b.z - LEN; d.z <= b.z + LEN2; d.z += LEN2)
		for (d.y = b.y - LEN; d.y <= b.y + LEN2; d.y += LEN2)
		for (d.x = b.x - LEN; d.x <= b.x + LEN2; d.x += LEN2) {
				pts[i] = this.newPoint(d);
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
				++i;
			}
		if (out == 8)
			return false;
		/* Only draw those faces that are visible */
		d.x = this.v.x - b.x;
		d.y = this.v.y - b.y;
		d.z = this.v.z - b.z;
		/*
			6----7
		       /    /|
		      /    / |
		     2----3  |
		    |  4 |  5
		    |    | /
		    0____1/
		*/
		if (d.z > LEN) {
			return inQuad(pt, pts, 4, 5, 7, 6);
		} else if (d.z < -LEN) {
			return inQuad(pt, pts, 0, 1, 3, 2);
		}
		// Could be overlap, need weighting then
		/*if (d.x > LEN) {
			return inQuad(pt, pts, 1, 3, 7, 5);
		} else if (d.x < -LEN) {
			return inQuad(pt, pts, 0, 2, 6, 4);
		}
		if (d.y > LEN) {
			return inQuad(pt, pts, 2, 3, 7, 6);
		} else if (d.y < -LEN) {
			return inQuad(pt, pts, 0, 1, 5, 4);
		}*/
		return false;
	}

	void drawCube(boolean w, Color[] cid, Color bid, Point3D cell) {
		int i = 0, out = 0;
		RealPoint d = new RealPoint();
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y, cell.z - this.o.z);

		for (d.z = b.z - LEN; d.z <= b.z + LEN2; d.z += LEN2)
		for (d.y = b.y - LEN; d.y <= b.y + LEN2; d.y += LEN2)
		for (d.x = b.x - LEN; d.x <= b.x + LEN2; d.x += LEN2) {
				pts[i] = this.newPoint(d);
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
				++i;
			}
		if (out == 8)
			return;
		/* Only draw those faces that are visible */
		d.x = this.v.x - b.x;
		d.y = this.v.y - b.y;
		d.z = this.v.z - b.z;
		/*
			6----7
		       /    /|
		      /    / |
		     2----3  |
		    |  4 |  5
		    |    | /
		    0____1/
		*/
		if (w || (cid[0] == null && cid[1] == null && cid[2] == null)) {
			if (d.z <= LEN)
				drawQuad(w, cid[2], bid, pts, 4, 5, 7, 6);
			else if (d.z >= -LEN)
				drawQuad(w, cid[2], bid, pts, 0, 1, 3, 2);
			if (d.x <= LEN)
				drawQuad(w, cid[0], bid, pts, 1, 3, 7, 5);
			else if (d.x >= -LEN)
				drawQuad(w, cid[0], bid, pts, 0, 2, 6, 4);
			if (d.y <= LEN)
				drawQuad(w, cid[1], bid, pts, 2, 3, 7, 6);
			else if (d.y >= -LEN)
				drawQuad(w, cid[1], bid, pts, 0, 1, 5, 4);
		}
		if (d.z > LEN)
			drawQuad(w, cid[2], bid, pts, 4, 5, 7, 6);
		else if (d.z < -LEN)
			drawQuad(w, cid[2], bid, pts, 0, 1, 3, 2);
		if (d.x > LEN)
			drawQuad(w, cid[0], bid, pts, 1, 3, 7, 5);
		else if (d.x < -LEN)
			drawQuad(w, cid[0], bid, pts, 0, 2, 6, 4);
		if (d.y > LEN)
			drawQuad(w, cid[1], bid, pts, 2, 3, 7, 6);
		else if (d.y < -LEN)
			drawQuad(w, cid[1], bid, pts, 0, 1, 5, 4);
	}

	void drawCubeMouth(boolean w, Color[] cid, Color bid, Point3D cell, short dir) {
		int i = 0, out = 0;
		RealPoint d = new RealPoint();
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y, cell.z - this.o.z);

		for (d.z = b.z - LEN; d.z <= b.z + LEN2; d.z += LEN2)
		for (d.y = b.y - LEN; d.y <= b.y + LEN2; d.y += LEN2)
		for (d.x = b.x - LEN; d.x <= b.x + LEN2; d.x += LEN2) {
				pts[i] = this.newPoint(d);
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
				++i;
			}
		if (out == 8)
			return;
		/* Only draw those faces that are visible */
		d.x = this.v.x - b.x;
		d.y = this.v.y - b.y;
		d.z = this.v.z - b.z;
		/*
			6----7
		       /    /|
		      /    / |
		     2----3  |
		    |  4 |  5
		    |    | /
		    0____1/
		*/
		if (w || (cid[0] == null && cid[1] == null && cid[2] == null)) {
			if (d.z <= LEN)
				drawQuad(w, cid[2], bid, pts, 4, 5, 7, 6);
			else if (d.z >= -LEN)
				drawQuad(w, cid[2], bid, pts, 0, 1, 3, 2);
			if (d.x <= LEN)
				drawQuad(w, cid[0], bid, pts, 1, 3, 7, 5);
			else if (d.x >= -LEN)
				drawQuad(w, cid[0], bid, pts, 0, 2, 6, 4);
			if (d.y <= LEN)
				drawQuad(w, cid[1], bid, pts, 2, 3, 7, 6);
			else if (d.y >= -LEN)
				drawQuad(w, cid[1], bid, pts, 0, 1, 5, 4);
		}
		if (d.z > LEN)
			drawQuadMouth(w, cid[2], bid, pts, 4, 5, 7, 6, dir == ZP);
		else if (d.z < -LEN)
			drawQuadMouth(w, cid[2], bid, pts, 0, 1, 3, 2, dir == ZN);
		if (d.x > LEN)
			drawQuadMouth(w, cid[0], bid, pts, 1, 3, 7, 5, dir == XP);
		else if (d.x < -LEN)
			drawQuadMouth(w, cid[0], bid, pts, 0, 2, 6, 4, dir == XN);
		if (d.y > LEN)
			drawQuadMouth(w, cid[1], bid, pts, 2, 3, 7, 6, dir == YP);
		else if (d.y < -LEN)
			drawQuadMouth(w, cid[1], bid, pts, 0, 1, 5, 4, dir == YN);
	}

	boolean inOctahedron(Point pt, Point3D cell) {
		int i, out = 0;
		RealPoint d = new RealPoint();
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y, (cell.z - this.o.z) / 2.0);

		for (int da = 0; da < 2; da++) {
			if (da == 1) {
				i = 0;
			} else {
				i = 3;
			}
			for (int db = 0; db < 3; db++) {
				if (db == 0) {
					d.x = (((da == 0) ? -LEN4 : LEN4) + b.x) / 2;
					d.y = b.y / 2;
					d.z = b.z / 2;
				} else if (db == 1) {
					d.y = (((da == 0) ? -LEN4 : LEN4) + b.y) / 2;
					d.z = b.z / 2;
					d.x = b.x / 2;
				} else {
					d.z = (((da == 0) ? -LEN4 : LEN4) + b.z) / 2;
					d.x = b.x / 2;
					d.y = b.y / 2;
				}
				pts[i] = this.newPoint(d);
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
				++i;
			}
		}
		if (out == 6)
			return false;

		/* Only draw those faces that are visible */
		d.x = this.v.x - b.x;
		d.y = this.v.y - b.y;
		d.z = this.v.z - b.z;
		/* back points looking through front
		       1

		   3   2   0

		       4
		  ----------
		   front points
		       1

		   3   5   0

		       4
		*/
		if (d.z >= -LEN) {
			if (inTri(pt, pts, 0, 1, 2))
				return true;
			if (inTri(pt, pts, 0, 2, 4))
				return true;
			if (inTri(pt, pts, 1, 3, 2))
				return true;
			if (inTri(pt, pts, 2, 3, 4))
				return true;
		} else if (d.z <= LEN) {
			if (inTri(pt, pts, 0, 4, 5))
				return true;
			if (inTri(pt, pts, 0, 5, 1))
				return true;
			if (inTri(pt, pts, 1, 5, 3))
				return true;
			if (inTri(pt, pts, 3, 5, 4))
				return true;
		}
		return false;
	}

	void drawOctahedron(boolean w, Color[] cid, Color bid, Point3D cell) {
		int i, out = 0;
		RealPoint d = new RealPoint();
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y, cell.z - this.o.z);

		for (int da = 0; da < 2; da++) {
			if (da == 1) {
				i = 0;
			} else {
				i = 3;
			}
			for (int db = 0; db < 3; db++) {
				if (db == 0) {
					d.x = (((da == 0) ? -LEN4 : LEN4) + b.x) / 2;
					d.y = b.y / 2;
					d.z = b.z / 2;
				} else if (db == 1) {
					d.y = (((da == 0) ? -LEN4 : LEN4) + b.y) / 2;
					d.z = b.z / 2;
					d.x = b.x / 2;
				} else {
					d.z = (((da == 0) ? -LEN4 : LEN4) + b.z) / 2;
					d.x = b.x / 2;
					d.y = b.y / 2;
				}
				pts[i] = this.newPoint(d);
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
				++i;
			}
		}
		if (out == 6)
			return;

		/* Only draw those faces that are visible */
		d.x = this.v.x - b.x;
		d.y = this.v.y - b.y;
		d.z = this.v.z - b.z;
		/* back points looking through front
		       1

		   3   2   0

		       4
		  ----------
		   front points
		       1

		   3   5   0

		       4
		*/
		if (w || (cid[0] == null && cid[1] == null && cid[2] == null)) {
			if (d.x + d.y + d.z < 0) {
				drawTri(w, cid[3], bid, pts, 2, 0, 1);
			} else {
				drawTri(w, cid[3], bid, pts, 5, 4, 3);
			}
			if (d.x + d.y - d.z < 0) {
				drawTri(w, cid[1], bid, pts, 5, 1, 0);
			} else {
				drawTri(w, cid[1], bid, pts, 2, 3, 4);
			}
			if (d.x - d.y + d.z < 0) {
				drawTri(w, cid[0], bid, pts, 2, 4, 0);
			} else {
				drawTri(w, cid[0], bid, pts, 5, 3, 1);
			}
			if (-d.x + d.y + d.z < 0) {
				drawTri(w, cid[2], bid, pts, 2, 1, 3);
			} else {
				drawTri(w, cid[2], bid, pts, 5, 0, 4);
			}
		}
		/*if (w || (cid[0] == null && cid[1] == null && cid[2] == null)) {
		}*/
		if (d.x + d.y + d.z >= 0) {
			drawTri(w, cid[3], bid, pts, 2, 0, 1);
		} else {
			drawTri(w, cid[3], bid, pts, 5, 4, 3);
		}
		if (d.x + d.y - d.z >= 0) {
			drawTri(w, cid[1], bid, pts, 5, 1, 0);
		} else {
			drawTri(w, cid[1], bid, pts, 2, 3, 4);
		}
		if (d.x - d.y + d.z >= 0) {
			drawTri(w, cid[0], bid, pts, 2, 4, 0);
		} else {
			drawTri(w, cid[0], bid, pts, 5, 3, 1);
		}
		if (-d.x + d.y + d.z >= 0) {
			drawTri(w, cid[2], bid, pts, 2, 1, 3);
		} else {
			drawTri(w, cid[2], bid, pts, 5, 0, 4);
		}
	}

	boolean inTetrahedronTrBl(Point pt, Point3D cell) {
		int i = 0, out = 0;
		RealPoint d = new RealPoint();
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y,
			cell.z - this.o.z);

		pts[i] = this.newPoint(d);
		d.z = (b.z - LEN2) / 2;
		d.y = (b.y - LEN2) / 2;
		d.x = (b.x + LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		d.z = (b.z - LEN2) / 2;
		d.y = (b.y + LEN2) / 2;
		d.x = (b.x - LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		d.z = (b.z + LEN2) / 2;
		d.y = (b.y - LEN2) / 2;
		d.x = (b.x - LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		d.z = (b.z + LEN2) / 2;
		d.y = (b.y + LEN2) / 2;
		d.x = (b.x + LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		if (out == 4)
			return false;
		/* Only draw those faces that are visible */
		d.x = this.v.x - b.x;
		d.y = this.v.y - b.y;
		d.z = this.v.z - b.z;
		/* back points looking through front
		       3

		   2
		  ----------
		   front points
		   1

		       0
		*/
		if (d.z >= -LEN) {
			if (inTri(pt, pts, 2, 3, 1))
				return true;
			if (inTri(pt, pts, 3, 2, 0))
				return true;
		} else if (d.z <= LEN) {
			if (inTri(pt, pts, 1, 0, 2))
				return true;
			if (inTri(pt, pts, 0, 1, 3))
				return true;
		}
		return false;
	}

	void drawTetrahedronTrBl(boolean w, Color[] cid, Color bid, Point3D cell) {
		int i = 0, out = 0;
		RealPoint d = new RealPoint();
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y,
			cell.z - this.o.z);

		pts[i] = this.newPoint(d);
		d.z = (b.z - LEN2) / 2;
		d.y = (b.y - LEN2) / 2;
		d.x = (b.x + LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		d.z = (b.z - LEN2) / 2;
		d.y = (b.y + LEN2) / 2;
		d.x = (b.x - LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		d.z = (b.z + LEN2) / 2;
		d.y = (b.y - LEN2) / 2;
		d.x = (b.x - LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		d.z = (b.z + LEN2) / 2;
		d.y = (b.y + LEN2) / 2;
		d.x = (b.x + LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		if (out == 4)
			return;
		/* Only draw those faces that are visible */
		d.x = this.v.x - b.x;
		d.y = this.v.y - b.y;
		d.z = this.v.z - b.z;
		/* back points looking through front
		       3

		   2
		  ----------
		   front points
		   1

		       0
		*/
		if (w || (cid[0] == null && cid[1] == null && cid[2] == null)) {
			if (d.x + d.y + d.z > 0) {
				drawTri(w, cid[3], bid, pts, 1, 0, 2);
			}
			if (d.x + d.y - d.z < 0) {
				drawTri(w, cid[1], bid, pts, 1, 3, 0);
			}
			if (d.x - d.y + d.z < 0) {
				drawTri(w, cid[0], bid, pts, 0, 2, 3);
			}
			if (-d.x + d.y + d.z < 0) {
				drawTri(w, cid[2], bid, pts, 1, 3, 2);
			}
		}
		if (d.x + d.y + d.z <= 0) {
			drawTri(w, cid[3], bid, pts, 1, 0, 2);
		}
		if (d.x + d.y - d.z >= 0) {
			drawTri(w, cid[1], bid, pts, 1, 3, 0);
		}
		if (d.x - d.y + d.z >= 0) {
			drawTri(w, cid[0], bid, pts, 0, 2, 3);
		}
		if (-d.x + d.y + d.z >= 0) {
			drawTri(w, cid[2], bid, pts, 1, 3, 2);
		}
	}

	boolean inTetrahedronTlBr(Point pt, Point3D cell) {
		int i = 0, out = 0;
		RealPoint d = new RealPoint();
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y,
			cell.z - this.o.z);

		pts[i] = this.newPoint(d);
		d.z = (b.z - LEN2) / 2;
		d.y = (b.y - LEN2) / 2;
		d.x = (b.x - LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		d.z = (b.z - LEN2) / 2;
		d.y = (b.y + LEN2) / 2;
		d.x = (b.x + LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		d.z = (b.z + LEN2) / 2;
		d.y = (b.y - LEN2) / 2;
		d.x = (b.x + LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		d.z = (b.z + LEN2) / 2;
		d.y = (b.y + LEN2) / 2;
		d.x = (b.x - LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		if (out == 4)
			return false;
		/* Only draw those faces that are visible */
		d.x = this.v.x - b.x;
		d.y = this.v.y - b.y;
		d.z = this.v.z - b.z;
		/* back points looking through front
		   1

		       0
		  ----------
		   front points
		       3

		   2
		*/
		if (d.z >= -LEN) {
			if (inTri(pt, pts, 0, 1, 2))
				return true;
			if (inTri(pt, pts, 1, 0, 3))
				return true;
		} else if (d.z <= LEN) {
			if (inTri(pt, pts, 3, 2, 1))
				return true;
			if (inTri(pt, pts, 2, 3, 0))
				return true;
		}
		return false;
	}

	void drawTetrahedronTlBr(boolean w, Color[] cid, Color bid, Point3D cell) {
		int i = 0, out = 0;
		RealPoint d = new RealPoint();
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y,
			cell.z - this.o.z);

		pts[i] = this.newPoint(d);
		d.z = (b.z - LEN2) / 2;
		d.y = (b.y - LEN2) / 2;
		d.x = (b.x - LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		d.z = (b.z - LEN2) / 2;
		d.y = (b.y + LEN2) / 2;
		d.x = (b.x + LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		d.z = (b.z + LEN2) / 2;
		d.y = (b.y - LEN2) / 2;
		d.x = (b.x + LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		d.z = (b.z + LEN2) / 2;
		d.y = (b.y + LEN2) / 2;
		d.x = (b.x - LEN2) / 2;
		pts[i] = this.newPoint(d);
		if (pts[i].x < 0 || pts[i].x >= this.width ||
				pts[i].y < 0 || pts[i].y >= this.height)
			++out;
		++i;
		if (out == 4)
			return;
		/* Only draw those faces that are visible */
		d.x = this.v.x - b.x;
		d.y = this.v.y - b.y;
		d.z = this.v.z - b.z;
		/* back points looking through front
		   1

		       0
		  ----------
		   front points
		       3

		   2
		*/
		if (w || (cid[0] == null && cid[1] == null && cid[2] == null)) {
			if (d.x + d.y + d.z > 0) {
				drawTri(w, cid[0], bid, pts, 1, 0, 2);
			}
			if (d.x + d.y - d.z < 0) {
				drawTri(w, cid[2], bid, pts, 1, 3, 0);
			}
			if (d.x - d.y + d.z < 0) {
				drawTri(w, cid[1], bid, pts, 0, 2, 3);
			}
			if (-d.x + d.y + d.z < 0) {
				drawTri(w, cid[3], bid, pts, 1, 3, 2);
			}
		}
		if (d.x + d.y + d.z <= 0) {
			drawTri(w, cid[0], bid, pts, 1, 0, 2);
		}
		if (d.x + d.y - d.z >= 0) {
			drawTri(w, cid[2], bid, pts, 1, 3, 0);
		}
		if (d.x - d.y + d.z >= 0) {
			drawTri(w, cid[1], bid, pts, 0, 2, 3);
		}
		if (-d.x + d.y + d.z >= 0) {
			drawTri(w, cid[3], bid, pts, 1, 3, 2);
		}
		/* if both 1st and 2nd... something wrong, kludge */
		if (d.x + d.y - d.z >= 0 && d.x - d.y + d.z >= 0) {
			drawTri(w, cid[0], bid, pts, 1, 0, 2);
			drawTri(w, cid[3], bid, pts, 1, 3, 2);
		}
	}

	boolean inRhombicDodecahedron(Point pt, Point3D cell) {
		int i = 3, out = 0;
		RealPoint d = new RealPoint();
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y, cell.z - this.o.z);

		/* Cube points first */
		for (d.x = b.x + LEN; d.x >= b.x - LEN2; d.x -= LEN2)
		for (d.y = b.y + LEN; d.y >= b.y - LEN2; d.y -= LEN2)
		for (d.z = b.z + LEN; d.z >= b.z - LEN2; d.z -= LEN2) {
				pts[i] = this.newPoint(d);
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
				++i;
		}
		/* Octahedron points */
		for (int da = 0; da < 2; da++) {
			if (da == 1) {
				i = 0;
			} else {
				i = 11; /* 3 + 8 */
			}
			for (int db = 0; db < 3; db++) {
				if (db == 0) {
					d.x = ((da == 0) ? -LEN2 : LEN2) + b.x;
					d.y = b.y;
					d.z = b.z;
				} else if (db == 1) {
					d.y = ((da == 0) ? -LEN2 : LEN2) + b.y;
					d.z = b.z;
					d.x = b.x;
				} else {
					d.z = ((da == 0) ? -LEN2 : LEN2) + b.z;
					d.x = b.x;
					d.y = b.y;
				}
				pts[i] = this.newPoint(d);
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
				++i;
			}
		}
		if (out == 14)
			return false;

		/* Only draw those faces that are visible */
		d.x = this.v.x - b.x;
		d.y = this.v.y - b.y;
		d.z = this.v.z - b.z;
		/* back points looking through front
		   8   1   4
		     7   3
		  11   2   0
		     9   5
		  10  12   6
		  ----------
		   front points
		   7   1   3
		     8   4
		  11  13   0
		    10   6
		   9  12   5
		*/
		if (d.z >= -LEN) {
			if (inQuad(pt, pts, 1, 7, 2, 3))
				return true;
			if (inQuad(pt, pts, 2, 7, 11, 9))
				return true;
			if (inQuad(pt, pts, 2, 9, 12, 5))
				return true;
			if (inQuad(pt, pts, 0, 3, 2, 5))
				return true;
		} else if (d.z <= LEN) {
			if (inQuad(pt, pts, 1, 4, 13, 8))
				return true;
			if (inQuad(pt, pts, 6, 12, 10, 13))
				return true;
			if (inQuad(pt, pts, 0, 6, 13, 4))
				return true;
			if (inQuad(pt, pts, 8, 13, 10, 11))
				return true;
		}
		/*	if (d.x <= LEN) {
			drawQuad(w, cid[1], bid, pts, 0, 3, 2, 5);
			drawQuad(w, cid[2], bid, pts, 0, 6, 13, 4);
			drawQuad(w, cid[1], bid, pts, 0, 4, 1, 3);
			drawQuad(w, cid[2], bid, pts, 0, 5, 12, 6);
			} else if (d.x >= -LEN) {
			drawQuad(w, cid[2], bid, pts, 2, 7, 11, 9);
			drawQuad(w, cid[1], bid, pts, 1, 8, 11, 7);
			drawQuad(w, cid[0], bid, pts, 8, 13, 10, 11);
			drawQuad(w, cid[1], bid, pts, 9, 11, 10, 12);
			}
			if (d.y <= LEN) {
			drawQuad(w, cid[1], bid, pts, 0, 4, 1, 3);
			drawQuad(w, cid[1], bid, pts, 1, 8, 11, 7);
			drawQuad(w, cid[2], bid, pts, 1, 7, 2, 3);
			drawQuad(w, cid[2], bid, pts, 1, 4, 13, 8);
			} else if (d.y >= -LEN) {
			drawQuad(w, cid[2], bid, pts, 0, 5, 12, 6);
			drawQuad(w, cid[1], bid, pts, 9, 11, 10, 12);
			drawQuad(w, cid[0], bid, pts, 2, 9, 12, 5);
			drawQuad(w, cid[1], bid, pts, 6, 12, 10, 13);
			}*/
		return false;
	}

	void drawRhombicDodecahedron(boolean w, Color[] cid, Color bid,
			Point3D cell) {
		int i = 3, out = 0, sortx, sorty, sortz;
		RealPoint d = new RealPoint();
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y, cell.z - this.o.z);

		/* Cube points first */
		for (d.x = b.x + LEN; d.x >= b.x - LEN2; d.x -= LEN2)
		for (d.y = b.y + LEN; d.y >= b.y - LEN2; d.y -= LEN2)
		for (d.z = b.z + LEN; d.z >= b.z - LEN2; d.z -= LEN2) {
				pts[i] = this.newPoint(d);
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
				++i;
		}
		/* Octahedron points */
		for (int da = 0; da < 2; da++) {
			if (da == 1) {
				i = 0;
			} else {
				i = 11; /* 3 + 8 */
			}
			for (int db = 0; db < 3; db++) {
				if (db == 0) {
					d.x = ((da == 0) ? -LEN2 : LEN2) + b.x;
					d.y = b.y;
					d.z = b.z;
				} else if (db == 1) {
					d.y = ((da == 0) ? -LEN2 : LEN2) + b.y;
					d.z = b.z;
					d.x = b.x;
				} else {
					d.z = ((da == 0) ? -LEN2 : LEN2) + b.z;
					d.x = b.x;
					d.y = b.y;
				}
				pts[i] = this.newPoint(d);
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
				++i;
			}
		}
		if (out == 14)
			return;

		/* Only draw those faces that are visible */
		d.x = this.v.x - b.x;
		d.y = this.v.y - b.y;
		d.z = this.v.z - b.z;
		/* back points looking through front
		   8   1   4
		     7   3
		  11   2   0
		     9   5
		  10  12   6
		  ----------
		   front points
		   7   1   3
		     8   4
		  11  13   0
		    10   6
		   9  12   5
		*/
		if (w || (cid[0] == null && cid[1] == null && cid[2] == null)) {
			if (d.x <= LEN) {
			drawQuad(w, cid[1], bid, pts, 0, 3, 2, 5);
			drawQuad(w, cid[2], bid, pts, 0, 6, 13, 4);
			drawQuad(w, cid[1], bid, pts, 0, 4, 1, 3);
			drawQuad(w, cid[2], bid, pts, 0, 5, 12, 6);
			} else if (d.x >= -LEN) {
			drawQuad(w, cid[2], bid, pts, 2, 7, 11, 9);
			drawQuad(w, cid[1], bid, pts, 1, 8, 11, 7);
			drawQuad(w, cid[0], bid, pts, 8, 13, 10, 11);
			drawQuad(w, cid[1], bid, pts, 9, 11, 10, 12);
			}
			if (d.y <= LEN) {
			drawQuad(w, cid[1], bid, pts, 0, 4, 1, 3);
			drawQuad(w, cid[1], bid, pts, 1, 8, 11, 7);
			drawQuad(w, cid[2], bid, pts, 1, 7, 2, 3);
			drawQuad(w, cid[2], bid, pts, 1, 4, 13, 8);
			} else if (d.y >= -LEN) {
			drawQuad(w, cid[2], bid, pts, 0, 5, 12, 6);
			drawQuad(w, cid[1], bid, pts, 9, 11, 10, 12);
			drawQuad(w, cid[0], bid, pts, 2, 9, 12, 5);
			drawQuad(w, cid[1], bid, pts, 6, 12, 10, 13);
			}
			if (d.z <= LEN) {
			drawQuad(w, cid[2], bid, pts, 1, 7, 2, 3);
			drawQuad(w, cid[0], bid, pts, 2, 9, 12, 5);
			drawQuad(w, cid[2], bid, pts, 2, 7, 11, 9);
			drawQuad(w, cid[1], bid, pts, 0, 3, 2, 5);
			} else if (d.z >= -LEN) {
			drawQuad(w, cid[2], bid, pts, 1, 4, 13, 8);
			drawQuad(w, cid[1], bid, pts, 6, 12, 10, 13);
			drawQuad(w, cid[2], bid, pts, 0, 6, 13, 4);
			drawQuad(w, cid[0], bid, pts, 8, 13, 10, 11);
			}
		}
		sortx = sorty = sortz = 0;
		if (java.lang.Math.abs(d.z - LEN) > java.lang.Math.abs(d.y - LEN))
			sortz++;
		if (java.lang.Math.abs(d.z - LEN) > java.lang.Math.abs(d.x - LEN))
			sortz++;
		if (java.lang.Math.abs(d.y - LEN) > java.lang.Math.abs(d.x - LEN))
			sorty++;
		if (java.lang.Math.abs(d.y - LEN) > java.lang.Math.abs(d.z - LEN))
			sorty++;
		if (java.lang.Math.abs(d.x - LEN) > java.lang.Math.abs(d.z - LEN))
			sortx++;
		if (java.lang.Math.abs(d.x - LEN) > java.lang.Math.abs(d.y - LEN))
			sortx++;
		for (i = 0; i < 3; i++) {
		  if (sortx == i) {
		    if (d.x > LEN) {
		      if (sortz > sorty) {
			if (d.z < 0)
			  drawQuad(w, cid[1], bid, pts, 0, 3, 2, 5);
			else
			  drawQuad(w, cid[2], bid, pts, 0, 6, 13, 4);
			drawQuad(w, cid[2], bid, pts, 0, 5, 12, 6);
			drawQuad(w, cid[1], bid, pts, 0, 4, 1, 3);
			if (d.z < 0)
			  drawQuad(w, cid[2], bid, pts, 0, 6, 13, 4);
			else
			  drawQuad(w, cid[1], bid, pts, 0, 3, 2, 5);
		      } else {
			if (d.y < 0)
			  drawQuad(w, cid[1], bid, pts, 0, 4, 1, 3);
			else
			  drawQuad(w, cid[2], bid, pts, 0, 5, 12, 6);
			drawQuad(w, cid[2], bid, pts, 0, 6, 13, 4);
			drawQuad(w, cid[1], bid, pts, 0, 3, 2, 5);
			if (d.y < 0)
			  drawQuad(w, cid[2], bid, pts, 0, 5, 12, 6);
			else
			  drawQuad(w, cid[1], bid, pts, 0, 4, 1, 3);
		      }
		    } else if (d.x < -LEN) {
		      if (sortz > sorty) {
			if (d.z < 0)
			  drawQuad(w, cid[2], bid, pts, 2, 7, 11, 9);
			else
			  drawQuad(w, cid[0], bid, pts, 8, 13, 10, 11);
			drawQuad(w, cid[1], bid, pts, 9, 11, 10, 12);
			drawQuad(w, cid[1], bid, pts, 1, 8, 11, 7);
			if (d.z < 0)
			  drawQuad(w, cid[2], bid, pts, 8, 13, 10, 11);
			else
			  drawQuad(w, cid[2], bid, pts, 2, 7, 11, 9);
		      } else {
			if (d.y < 0)
			  drawQuad(w, cid[1], bid, pts, 1, 8, 11, 7);
			else
			  drawQuad(w, cid[1], bid, pts, 9, 11, 10, 12);
			drawQuad(w, cid[0], bid, pts, 8, 13, 10, 11);
			drawQuad(w, cid[2], bid, pts, 2, 7, 11, 9);
			if (d.y < 0)
			  drawQuad(w, cid[1], bid, pts, 9, 11, 10, 12);
			else
			  drawQuad(w, cid[1], bid, pts, 1, 8, 11, 7);
		      }
		    }
		  }
		  if (sorty == i) {
		    if (d.y > LEN) {
		      if (sortx > sortz) {
			if (d.x < 0)
			  drawQuad(w, cid[1], bid, pts, 0, 4, 1, 3);
			else
			  drawQuad(w, cid[1], bid, pts, 1, 8, 11, 7);
			drawQuad(w, cid[0], bid, pts, 1, 4, 13, 8);
			drawQuad(w, cid[2], bid, pts, 1, 7, 2, 3);
			if (d.x < 0)
			  drawQuad(w, cid[1], bid, pts, 1, 8, 11, 7);
			else
			  drawQuad(w, cid[1], bid, pts, 0, 4, 1, 3);
		      } else {
			if (d.z < 0)
			  drawQuad(w, cid[2], bid, pts, 1, 7, 2, 3);
			else
			  drawQuad(w, cid[2], bid, pts, 1, 4, 13, 8);
			drawQuad(w, cid[1], bid, pts, 1, 8, 11, 7);
			drawQuad(w, cid[1], bid, pts, 0, 4, 1, 3);
			if (d.z < 0)
			  drawQuad(w, cid[2], bid, pts, 1, 4, 13, 8);
			else
			  drawQuad(w, cid[2], bid, pts, 1, 7, 2, 3);
		      }
		    } else if (d.y < -LEN) {
		      if (sortx > sortz) {
			if (d.x < 0)
			  drawQuad(w, cid[2], bid, pts, 0, 5, 12, 6);
			else
			  drawQuad(w, cid[1], bid, pts, 9, 11, 10, 12);
			drawQuad(w, cid[1], bid, pts, 6, 12, 10, 13);
			drawQuad(w, cid[0], bid, pts, 2, 9, 12, 5);
			if (d.x < 0)
			  drawQuad(w, cid[1], bid, pts, 9, 11, 10, 12);
			else
			  drawQuad(w, cid[2], bid, pts, 0, 5, 12, 6);
		      } else {
			if (d.z < 0)
			  drawQuad(w, cid[0], bid, pts, 2, 9, 12, 5);
			else
			  drawQuad(w, cid[1], bid, pts, 6, 12, 10, 13);
			drawQuad(w, cid[1], bid, pts, 9, 11, 10, 12);
			drawQuad(w, cid[2], bid, pts, 0, 5, 12, 6);
			if (d.z < 0)
			  drawQuad(w, cid[1], bid, pts, 6, 12, 10, 13);
			else
			  drawQuad(w, cid[0], bid, pts, 2, 9, 12, 5);
		      }
		    }
		  }
		  if (sortz == i) {
		    if (d.z > LEN) {
		      if (sorty > sortx) {
			if (d.y < 0)
			  drawQuad(w, cid[2], bid, pts, 1, 7, 2, 3);
			else
			  drawQuad(w, cid[0], bid, pts, 2, 9, 12, 5);
			drawQuad(w, cid[1], bid, pts, 0, 3, 2, 5);
			drawQuad(w, cid[2], bid, pts, 2, 7, 11, 9);
			if (d.y < 0)
			  drawQuad(w, cid[0], bid, pts, 2, 9, 12, 5);
			else
			  drawQuad(w, cid[2], bid, pts, 1, 7, 2, 3);
		      } else {
			if (d.x < 0)
			  drawQuad(w, cid[1], bid, pts, 0, 3, 2, 5);
			else
			  drawQuad(w, cid[2], bid, pts, 2, 7, 11, 9);
			drawQuad(w, cid[0], bid, pts, 2, 9, 12, 5);
			drawQuad(w, cid[2], bid, pts, 1, 7, 2, 3);
			if (d.x < 0)
			  drawQuad(w, cid[2], bid, pts, 2, 7, 11, 9);
			else
			  drawQuad(w, cid[1], bid, pts, 0, 3, 2, 5);
		      }
		    } else if (d.z < -LEN) {
		      if (sorty > sortx) {
			if (d.y < 0)
			  drawQuad(w, cid[2], bid, pts, 1, 4, 13, 8);
			else
			  drawQuad(w, cid[1], bid, pts, 6, 12, 10, 13);
			drawQuad(w, cid[0], bid, pts, 8, 13, 10, 11);
			drawQuad(w, cid[2], bid, pts, 0, 6, 13, 4);
			if (d.y < 0)
			  drawQuad(w, cid[1], bid, pts, 6, 12, 10, 13);
			else
			  drawQuad(w, cid[2], bid, pts, 1, 4, 13, 8);
		      } else {
			if (d.x < 0)
			  drawQuad(w, cid[2], bid, pts, 0, 6, 13, 4);
			else
			  drawQuad(w, cid[0], bid, pts, 8, 13, 10, 11);
			drawQuad(w, cid[1], bid, pts, 6, 12, 10, 13);
			drawQuad(w, cid[2], bid, pts, 1, 4, 13, 8);
			if (d.x < 0)
			  drawQuad(w, cid[0], bid, pts, 8, 13, 10, 11);
			else
			  drawQuad(w, cid[2], bid, pts, 0, 6, 13, 4);
		      }
		    }
		  }
		}
	}

	void drawRhombicDodecahedronMouth(boolean w, Color[] cid, Color bid,
			Point3D cell, short dir) {
		int i = 3, out = 0, sortx, sorty, sortz;
		RealPoint d = new RealPoint();
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y, cell.z - this.o.z);

		/* Cube points first */
		for (d.x = b.x + LEN; d.x >= b.x - LEN2; d.x -= LEN2)
		for (d.y = b.y + LEN; d.y >= b.y - LEN2; d.y -= LEN2)
		for (d.z = b.z + LEN; d.z >= b.z - LEN2; d.z -= LEN2) {
				pts[i] = this.newPoint(d);
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
				++i;
		}
		/* Octahedron points */
		for (int da = 0; da < 2; da++) {
			if (da == 1) {
				i = 0;
			} else {
				i = 11; /* 3 + 8 */
			}
			for (int db = 0; db < 3; db++) {
				if (db == 0) {
					d.x = ((da == 0) ? -LEN2 : LEN2) + b.x;
					d.y = b.y;
					d.z = b.z;
				} else if (db == 1) {
					d.y = ((da == 0) ? -LEN2 : LEN2) + b.y;
					d.z = b.z;
					d.x = b.x;
				} else {
					d.z = ((da == 0) ? -LEN2 : LEN2) + b.z;
					d.x = b.x;
					d.y = b.y;
				}
				pts[i] = this.newPoint(d);
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
				++i;
			}
		}
		if (out == 14)
			return;

		/* Only draw those faces that are visible */
		d.x = this.v.x - b.x;
		d.y = this.v.y - b.y;
		d.z = this.v.z - b.z;
		/* back points looking through front
		   8   1   4
		     7   3
		  11   2   0
		     9   5
		  10  12   6
		  ----------
		   front points
		   7   1   3
		     8   4
		  11  13   0
		    10   6
		   9  12   5
		*/
		if (w || (cid[0] == null && cid[1] == null && cid[2] == null)) {
			if (d.x <= LEN) { // 0
			drawQuad(w, cid[1], bid, pts, 0, 3, 2, 5); /* ZNXN */
			drawQuad(w, cid[2], bid, pts, 0, 6, 13, 4);  /* ZPXN */
			drawQuad(w, cid[1], bid, pts, 0, 4, 1, 3); /* XNYN */
			drawQuad(w, cid[2], bid, pts, 0, 5, 12, 6); /* XNYP */
			} else if (d.x >= -LEN) { // 11
			drawQuad(w, cid[2], bid, pts, 2, 7, 11, 9); /* ZNXP */
			drawQuad(w, cid[1], bid, pts, 1, 8, 11, 7); /* XPYN */
			drawQuad(w, cid[0], bid, pts, 8, 13, 10, 11); /* ZPXP */
			drawQuad(w, cid[1], bid, pts, 9, 11, 10, 12); /* XPYP */
			}
			if (d.y <= LEN) { // 1
			drawQuad(w, cid[1], bid, pts, 0, 4, 1, 3); /* XNYN */
			drawQuad(w, cid[1], bid, pts, 1, 8, 11, 7); /* XPYN */
			drawQuad(w, cid[2], bid, pts, 1, 7, 2, 3); /* YNZN */
			drawQuad(w, cid[2], bid, pts, 1, 4, 13, 8); /* YNZP */
			} else if (d.y >= -LEN) { // 12
			drawQuad(w, cid[2], bid, pts, 0, 5, 12, 6); /* XNYP */
			drawQuad(w, cid[1], bid, pts, 9, 11, 10, 12); /* XPYP */
			drawQuad(w, cid[0], bid, pts, 2, 9, 12, 5); /* YPZN */
			drawQuad(w, cid[1], bid, pts, 6, 12, 10, 13); /* YPZP */
			}
			if (d.z <= LEN) { // 2
			drawQuad(w, cid[2], bid, pts, 1, 7, 2, 3); /* YNZN */
			drawQuad(w, cid[0], bid, pts, 2, 9, 12, 5); /* YPZN */
			drawQuad(w, cid[2], bid, pts, 2, 7, 11, 9); /* ZNXP */
			drawQuad(w, cid[1], bid, pts, 0, 3, 2, 5); /* ZNXN */
			} else if (d.z >= -LEN) { // 13
			drawQuad(w, cid[2], bid, pts, 1, 4, 13, 8); /* YNZP */
			drawQuad(w, cid[1], bid, pts, 6, 12, 10, 13); /* YPZP */
			drawQuad(w, cid[2], bid, pts, 0, 6, 13, 4); /* ZPXN */
			drawQuad(w, cid[0], bid, pts, 8, 13, 10, 11); /* ZPXP */
			}
		}
		sortx = sorty = sortz = 0;
		if (java.lang.Math.abs(d.z - LEN) > java.lang.Math.abs(d.y - LEN))
			sortz++;
		if (java.lang.Math.abs(d.z - LEN) > java.lang.Math.abs(d.x - LEN))
			sortz++;
		if (java.lang.Math.abs(d.y - LEN) > java.lang.Math.abs(d.x - LEN))
			sorty++;
		if (java.lang.Math.abs(d.y - LEN) > java.lang.Math.abs(d.z - LEN))
			sorty++;
		if (java.lang.Math.abs(d.x - LEN) > java.lang.Math.abs(d.z - LEN))
			sortx++;
		if (java.lang.Math.abs(d.x - LEN) > java.lang.Math.abs(d.y - LEN))
			sortx++;
		for (i = 0; i < 3; i++) {
		  if (sortx == i) {
		    if (d.x > LEN) {
		      if (sortz > sorty) {
			if (d.z < 0)
			  drawQuadMouth(w, cid[1], bid, pts, 0, 3, 2, 5, dir == ZPXP);
			else
			  drawQuadMouth(w, cid[2], bid, pts, 0, 6, 13, 4, dir == ZNXP);
			drawQuadMouth(w, cid[2], bid, pts, 0, 5, 12, 6, dir == XPYN);
			drawQuadMouth(w, cid[1], bid, pts, 0, 4, 1, 3, dir == XPYP);
			if (d.z < 0)
			  drawQuadMouth(w, cid[2], bid, pts, 0, 6, 13, 4, dir == ZNXP);
			else
			  drawQuadMouth(w, cid[1], bid, pts, 0, 3, 2, 5, dir == ZPXP);
		      } else {
			if (d.y < 0)
			  drawQuadMouth(w, cid[2], bid, pts, 0, 4, 1, 3, dir == XPYP);
			else
			  drawQuadMouth(w, cid[1], bid, pts, 0, 5, 12, 6, dir == XPYN);
			drawQuadMouth(w, cid[2], bid, pts, 0, 6, 13, 4, dir == ZNXP);
			drawQuadMouth(w, cid[1], bid, pts, 0, 3, 2, 5, dir == ZPXP);
			if (d.y < 0)
			  drawQuadMouth(w, cid[2], bid, pts, 0, 5, 12, 6, dir == XPYN);
			else
			  drawQuadMouth(w, cid[1], bid, pts, 0, 4, 1, 3, dir == XPYP);
		      }
		    } else if (d.x < -LEN) {
		      if (sortz > sorty) {
			if (d.z < 0)
			  drawQuadMouth(w, cid[2], bid, pts, 2, 7, 11, 9, dir == ZPXN);
			else
			  drawQuadMouth(w, cid[0], bid, pts, 8, 13, 10, 11, dir == ZNXN);
			drawQuadMouth(w, cid[1], bid, pts, 9, 11, 10, 12, dir == XNYN);
			drawQuadMouth(w, cid[1], bid, pts, 1, 8, 11, 7, dir == XNYP);
			if (d.z < 0)
			  drawQuadMouth(w, cid[2], bid, pts, 8, 13, 10, 11, dir == ZNXN);
			else
			  drawQuadMouth(w, cid[2], bid, pts, 2, 7, 11, 9, dir == ZPXN);
		      } else {
			if (d.y < 0)
			  drawQuadMouth(w, cid[1], bid, pts, 1, 8, 11, 7, dir == XNYP);
			else
			  drawQuadMouth(w, cid[1], bid, pts, 9, 11, 10, 12, dir == XNYN);
			drawQuadMouth(w, cid[0], bid, pts, 8, 13, 10, 11, dir == ZNXN);
			drawQuadMouth(w, cid[2], bid, pts, 2, 7, 11, 9, dir == ZPXN);
			if (d.y < 0)
			  drawQuadMouth(w, cid[1], bid, pts, 9, 11, 10, 12, dir == XNYN);
			else
			  drawQuadMouth(w, cid[1], bid, pts, 1, 8, 11, 7, dir == XNYP);
		      }
		    }
		  }
		  if (sorty == i) {
		    if (d.y > LEN) {
		      if (sortx > sortz) {
			if (d.x < 0)
			  drawQuadMouth(w, cid[1], bid, pts, 0, 4, 1, 3, dir == XPYP);
			else
			  drawQuadMouth(w, cid[1], bid, pts, 1, 8, 11, 7, dir == XNYP);
			drawQuadMouth(w, cid[0], bid, pts, 1, 4, 13, 8, dir == YPZN);
			drawQuadMouth(w, cid[2], bid, pts, 1, 7, 2, 3, dir == YPZP);
			if (d.x < 0)
			  drawQuadMouth(w, cid[1], bid, pts, 1, 8, 11, 7, dir == XNYP);
			else
			  drawQuadMouth(w, cid[1], bid, pts, 0, 4, 1, 3, dir == XPYP);
		      } else {
			if (d.z < 0)
			  drawQuadMouth(w, cid[2], bid, pts, 1, 7, 2, 3, dir == YPZP);
			else
			  drawQuadMouth(w, cid[2], bid, pts, 1, 4, 13, 8, dir == YPZN);
			drawQuadMouth(w, cid[1], bid, pts, 1, 8, 11, 7, dir == XNYP);
			drawQuadMouth(w, cid[1], bid, pts, 0, 4, 1, 3, dir == XPYP);
			if (d.z < 0)
			  drawQuadMouth(w, cid[2], bid, pts, 1, 4, 13, 8, dir == YPZN);
			else
			  drawQuadMouth(w, cid[2], bid, pts, 1, 7, 2, 3, dir == YPZP);
		      }
		    } else if (d.y < -LEN) {
		      if (sortx > sortz) {
			if (d.x < 0)
			  drawQuadMouth(w, cid[2], bid, pts, 0, 5, 12, 6, dir == XPYN);
			else
			  drawQuadMouth(w, cid[1], bid, pts, 9, 11, 10, 12, dir == XNYN);
			drawQuadMouth(w, cid[1], bid, pts, 6, 12, 10, 13, dir == YNZP);
			drawQuadMouth(w, cid[0], bid, pts, 2, 9, 12, 5, dir == YNZN);
			if (d.x < 0)
			  drawQuadMouth(w, cid[1], bid, pts, 9, 11, 10, 12, dir == XNYN);
			else
			  drawQuadMouth(w, cid[2], bid, pts, 0, 5, 12, 6, dir == XPYN);
		      } else {
			if (d.z < 0)
			  drawQuadMouth(w, cid[0], bid, pts, 2, 9, 12, 5, dir == YNZP);
			else
			  drawQuadMouth(w, cid[1], bid, pts, 6, 12, 10, 13, dir == YNZN);
			drawQuadMouth(w, cid[1], bid, pts, 9, 11, 10, 12, dir == XNYN);
			drawQuadMouth(w, cid[2], bid, pts, 0, 5, 12, 6, dir == XPYN);
			if (d.z < 0)
			  drawQuadMouth(w, cid[1], bid, pts, 6, 12, 10, 13, dir == YNZN);
			else
			  drawQuadMouth(w, cid[0], bid, pts, 2, 9, 12, 5, dir == YNZP);
		      }
		    }
		  }
		  if (sortz == i) {
		    if (d.z > LEN) {
		      if (sorty > sortx) {
			if (d.y < 0)
			  drawQuadMouth(w, cid[2], bid, pts, 1, 7, 2, 3, dir == YPZP);
			else
			  drawQuadMouth(w, cid[0], bid, pts, 2, 9, 12, 5, dir == YNZP);
			drawQuadMouth(w, cid[1], bid, pts, 0, 3, 2, 5, dir == ZPXP);
			drawQuadMouth(w, cid[2], bid, pts, 2, 7, 11, 9, dir == ZPXN);
			if (d.y < 0)
			  drawQuadMouth(w, cid[0], bid, pts, 2, 9, 12, 5, dir == YNZP);
			else
			  drawQuadMouth(w, cid[2], bid, pts, 1, 7, 2, 3, dir == YPZP);
		      } else {
			if (d.x < 0)
			  drawQuadMouth(w, cid[1], bid, pts, 0, 3, 2, 5, dir == ZPXP);
			else
			  drawQuadMouth(w, cid[2], bid, pts, 2, 7, 11, 9, dir == ZPXN);
			drawQuadMouth(w, cid[0], bid, pts, 2, 9, 12, 5, dir == YNZP);
			drawQuadMouth(w, cid[2], bid, pts, 1, 7, 2, 3, dir == YPZP);
			if (d.x < 0)
			  drawQuadMouth(w, cid[2], bid, pts, 2, 7, 11, 9, dir == ZPXN);
			else
			  drawQuadMouth(w, cid[1], bid, pts, 0, 3, 2, 5, dir == ZPXP);
		      }
		    } else if (d.z < -LEN) {
		      if (sorty > sortx) {
			if (d.y < 0)
			  drawQuadMouth(w, cid[2], bid, pts, 1, 4, 13, 8, dir == YPZN);
			else
			  drawQuadMouth(w, cid[1], bid, pts, 6, 12, 10, 13, dir == YNZN);
			drawQuadMouth(w, cid[0], bid, pts, 8, 13, 10, 11, dir == ZNXN);
			drawQuadMouth(w, cid[2], bid, pts, 0, 6, 13, 4, dir == ZNXP);
			if (d.y < 0)
			  drawQuadMouth(w, cid[1], bid, pts, 6, 12, 10, 13, dir == YNZN);
			else
			  drawQuadMouth(w, cid[2], bid, pts, 1, 4, 13, 8, dir == YPZN);
		      } else {
			if (d.x < 0)
			  drawQuadMouth(w, cid[2], bid, pts, 0, 6, 13, 4, dir == ZNXP);
			else
			  drawQuadMouth(w, cid[0], bid, pts, 8, 13, 10, 11, dir == ZNXN);
			drawQuadMouth(w, cid[1], bid, pts, 6, 12, 10, 13, dir == YNZN);
			drawQuadMouth(w, cid[2], bid, pts, 1, 4, 13, 8, dir == YPZN);
			if (d.x < 0)
			  drawQuadMouth(w, cid[0], bid, pts, 8, 13, 10, 11, dir == ZNXN);
			else
			  drawQuadMouth(w, cid[2], bid, pts, 0, 6, 13, 4, dir == ZNXP);
		      }
		    }
		  }
		}
/*System.out.println("d:" + d.x + " " + d.y + " " + d.z + ", " +
  "s:" + sortx + " " + sorty + " " + sortz + ", " + printDir(12, dir));*/
	}

	boolean inTruncatedOctahedron(Point pt, Point3D cell) {
		int i = 0, out = 0;
		RealPoint d = new RealPoint();
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y, (cell.z - this.o.z) / 2.0);

		if ((cell.z & 1) == 1) {
			b.x += 0.5;
			b.y += 0.5;
		}
		pts[0] = this.newPoint(new RealPoint(b, LEN_2, 0., LEN));
		pts[1] = this.newPoint(new RealPoint(b, 0., LEN_2, LEN));
		pts[2] = this.newPoint(new RealPoint(b, -LEN_2, 0., LEN));
		pts[3] = this.newPoint(new RealPoint(b, 0., -LEN_2, LEN));

		pts[4] = this.newPoint(new RealPoint(b, LEN, 0., LEN_2));
		pts[15] = this.newPoint(new RealPoint(b, LEN, -LEN_2, 0.));
		pts[16] = this.newPoint(new RealPoint(b, LEN, 0., -LEN_2));
		pts[8] = this.newPoint(new RealPoint(b, LEN, LEN_2, 0.));

		pts[5] = this.newPoint(new RealPoint(b, 0., LEN, LEN_2));
		pts[9] = this.newPoint(new RealPoint(b, LEN_2, LEN, 0.));
		pts[17] = this.newPoint(new RealPoint(b, 0., LEN, -LEN_2));
		pts[10] = this.newPoint(new RealPoint(b, -LEN_2, LEN, 0.));

		pts[6] = this.newPoint(new RealPoint(b, -LEN, 0., LEN_2));
		pts[11] = this.newPoint(new RealPoint(b, -LEN, LEN_2, 0.));
		pts[18] = this.newPoint(new RealPoint(b, -LEN, 0., -LEN_2));
		pts[12] = this.newPoint(new RealPoint(b, -LEN, -LEN_2, 0.));

		pts[7] = this.newPoint(new RealPoint(b, 0., -LEN, LEN_2));
		pts[13] = this.newPoint(new RealPoint(b, -LEN_2, -LEN, 0.));
		pts[19] = this.newPoint(new RealPoint(b, 0., -LEN, -LEN_2));
		pts[14] = this.newPoint(new RealPoint(b, LEN_2, -LEN, 0.));

		pts[20] = this.newPoint(new RealPoint(b, LEN_2, 0., -LEN));
		pts[21] = this.newPoint(new RealPoint(b, 0., LEN_2, -LEN));
		pts[22] = this.newPoint(new RealPoint(b, -LEN_2, 0., -LEN));
		pts[23] = this.newPoint(new RealPoint(b, 0., -LEN_2, -LEN));
		for (i = 0; i < 24; i++) {
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
		}
		if (out == 24)
			return false;

		/* Only draw those faces that are visible */
		d.x = this.v.x - b.x;
		d.y = this.v.y - b.y;
		d.z = this.v.z - b.z;
		/* back points looking through front
		   10  5  9
		 11    1    8
		  6  2   0  4
		 12    3   15
		   13  7 14
		  ----------
		   front points
		   10 17  9
		 11   21    8
		 18 22  20 16
		 12   23   15
		   13 19 14
		*/
		if (d.z >= -LEN) {
			if (inQuad(pt, pts, 0, 1, 2, 3))
				return true;
			if (inHex(pt, pts, 1, 0, 4, 8, 9, 5))
				return true;
			if (inHex(pt, pts, 3, 2, 6, 12, 13, 7))
				return true;
			if (inHex(pt, pts, 0, 3, 7, 14, 15, 4))
				return true;
			if (inHex(pt, pts, 2, 1, 5, 10, 11, 6))
				return true;
		} else if (d.z <= LEN) {
			if (inQuad(pt, pts, 23, 22, 21, 20))
				return true;
			if (inHex(pt, pts, 22, 23, 19, 13, 12, 18))
				return true;
			if (inHex(pt, pts, 20, 21, 17, 9, 8, 16))
				return true;
			if (inHex(pt, pts, 21, 22, 18, 11, 10, 17))
				return true;
			if (inHex(pt, pts, 23, 20, 16, 15, 14, 19))
				return true;
		}
		return false;
	}

	void drawTruncatedOctahedron(boolean w, Color[] cid, Color bid,
			Point3D cell) {
		int i = 0, out = 0;
		RealPoint d = new RealPoint();
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y, (cell.z - this.o.z) / 2.0);

		if ((cell.z & 1) == 1) {
			b.x += 0.5;
			b.y += 0.5;
		}
		pts[0] = this.newPoint(new RealPoint(b, LEN_2, 0., LEN));
		pts[1] = this.newPoint(new RealPoint(b, 0., LEN_2, LEN));
		pts[2] = this.newPoint(new RealPoint(b, -LEN_2, 0., LEN));
		pts[3] = this.newPoint(new RealPoint(b, 0., -LEN_2, LEN));

		pts[4] = this.newPoint(new RealPoint(b, LEN, 0., LEN_2));
		pts[15] = this.newPoint(new RealPoint(b, LEN, -LEN_2, 0.));
		pts[16] = this.newPoint(new RealPoint(b, LEN, 0., -LEN_2));
		pts[8] = this.newPoint(new RealPoint(b, LEN, LEN_2, 0.));

		pts[5] = this.newPoint(new RealPoint(b, 0., LEN, LEN_2));
		pts[9] = this.newPoint(new RealPoint(b, LEN_2, LEN, 0.));
		pts[17] = this.newPoint(new RealPoint(b, 0., LEN, -LEN_2));
		pts[10] = this.newPoint(new RealPoint(b, -LEN_2, LEN, 0.));

		pts[6] = this.newPoint(new RealPoint(b, -LEN, 0., LEN_2));
		pts[11] = this.newPoint(new RealPoint(b, -LEN, LEN_2, 0.));
		pts[18] = this.newPoint(new RealPoint(b, -LEN, 0., -LEN_2));
		pts[12] = this.newPoint(new RealPoint(b, -LEN, -LEN_2, 0.));

		pts[7] = this.newPoint(new RealPoint(b, 0., -LEN, LEN_2));
		pts[13] = this.newPoint(new RealPoint(b, -LEN_2, -LEN, 0.));
		pts[19] = this.newPoint(new RealPoint(b, 0., -LEN, -LEN_2));
		pts[14] = this.newPoint(new RealPoint(b, LEN_2, -LEN, 0.));

		pts[20] = this.newPoint(new RealPoint(b, LEN_2, 0., -LEN));
		pts[21] = this.newPoint(new RealPoint(b, 0., LEN_2, -LEN));
		pts[22] = this.newPoint(new RealPoint(b, -LEN_2, 0., -LEN));
		pts[23] = this.newPoint(new RealPoint(b, 0., -LEN_2, -LEN));
		for (i = 0; i < 24; i++) {
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
		}
		if (out == 24)
			return;

		/* Only draw those faces that are visible */
		d.x = this.v.x - b.x;
		d.y = this.v.y - b.y;
		d.z = this.v.z - b.z;
		/* back points looking through front
		   10  5  9
		 11    1    8
		  6  2   0  4
		 12    3   15
		   13  7 14
		  ----------
		   front points
		   10 17  9
		 11   21    8
		 18 22  20 16
		 12   23   15
		   13 19 14
		*/
		if (w || (cid[0] == null && cid[1] == null && cid[2] == null)) {
			if (d.z <= LEN)
				drawQuad(w, cid[2], bid, pts, 0, 1, 2, 3);
			else if (d.z >= -LEN)
				drawQuad(w, cid[2], bid, pts, 23, 22, 21, 20);
			if (d.y <= LEN)
				drawQuad(w, cid[1], bid, pts, 5, 9, 17, 10);
			else if (d.y >= -LEN)
				drawQuad(w, cid[1], bid, pts, 7, 14, 19, 13);
			if (d.x <= LEN)
				drawQuad(w, cid[0], bid, pts, 4, 8, 16, 15);
			else if (d.x >= -LEN)
				drawQuad(w, cid[0], bid, pts, 6, 11, 18, 12);
			if (d.x + d.y + d.z < 0) {
				drawHex(w, cid[3], bid, pts, 1, 0, 4, 8, 9, 5);
			} else {
				drawHex(w, cid[3], bid, pts, 22, 23, 19, 13, 12, 18);
			}
			if (d.x + d.y - d.z < 0) {
				drawHex(w, cid[1], bid, pts, 20, 21, 17, 9, 8, 16);
			} else {
				drawHex(w, cid[1], bid, pts, 3, 2, 6, 12, 13, 7);
			}
			if (d.x - d.y + d.z < 0) {
				drawHex(w, cid[0], bid, pts, 0, 3, 7, 14, 15, 4);
			} else {
				drawHex(w, cid[0], bid, pts, 21, 22, 18, 11, 10, 17);
			}
			if (-d.x + d.y + d.z < 0) {
				drawHex(w, cid[2], bid, pts, 2, 1, 5, 10, 11, 6);
			} else {
				drawHex(w, cid[2], bid, pts, 23, 20, 16, 15, 14, 19);
			}
		}
		if (d.x + d.y + d.z >= 0) {
			drawHex(w, cid[3], bid, pts, 1, 0, 4, 8, 9, 5);
		} else {
			drawHex(w, cid[3], bid, pts, 22, 23, 19, 13, 12, 18);
		}
		if (d.x + d.y - d.z >= 0) {
			drawHex(w, cid[1], bid, pts, 20, 21, 17, 9, 8, 16);
		} else {
			drawHex(w, cid[1], bid, pts, 3, 2, 6, 12, 13, 7);
		}
		if (d.x - d.y + d.z >= 0) {
			drawHex(w, cid[0], bid, pts, 0, 3, 7, 14, 15, 4);
		} else {
			drawHex(w, cid[0], bid, pts, 21, 22, 18, 11, 10, 17);
		}
		if (-d.x + d.y + d.z >= 0) {
			drawHex(w, cid[2], bid, pts, 2, 1, 5, 10, 11, 6);
		} else {
			drawHex(w, cid[2], bid, pts, 23, 20, 16, 15, 14, 19);
		}
		if (d.z > LEN)
			drawQuad(w, cid[2], bid, pts, 0, 1, 2, 3);
		else if (d.z < -LEN)
			drawQuad(w, cid[2], bid, pts, 23, 22, 21, 20);
		if (d.y > LEN)
			drawQuad(w, cid[1], bid, pts, 5, 9, 17, 10);
		else if (d.y < -LEN)
			drawQuad(w, cid[1], bid, pts, 7, 14, 19, 13);
		if (d.x > LEN)
			drawQuad(w, cid[0], bid, pts, 4, 8, 16, 15);
		else if (d.x < -LEN)
			drawQuad(w, cid[0], bid, pts, 6, 11, 18, 12);
	}

	void drawTruncatedOctahedronMouth(boolean w, Color[] cid, Color bid,
			Point3D cell, short dir) {
		int i = 0, out = 0;
		RealPoint d = new RealPoint();
		RealPoint b = new RealPoint(cell.x - this.o.x,
			cell.y - this.o.y, (cell.z - this.o.z) / 2.0);

		if ((cell.z & 1) == 1) {
			b.x += 0.5;
			b.y += 0.5;
		}
		pts[0] = this.newPoint(new RealPoint(b, LEN_2, 0., LEN));
		pts[1] = this.newPoint(new RealPoint(b, 0., LEN_2, LEN));
		pts[2] = this.newPoint(new RealPoint(b, -LEN_2, 0., LEN));
		pts[3] = this.newPoint(new RealPoint(b, 0., -LEN_2, LEN));

		pts[4] = this.newPoint(new RealPoint(b, LEN, 0., LEN_2));
		pts[15] = this.newPoint(new RealPoint(b, LEN, -LEN_2, 0.));
		pts[16] = this.newPoint(new RealPoint(b, LEN, 0., -LEN_2));
		pts[8] = this.newPoint(new RealPoint(b, LEN, LEN_2, 0.));

		pts[5] = this.newPoint(new RealPoint(b, 0., LEN, LEN_2));
		pts[9] = this.newPoint(new RealPoint(b, LEN_2, LEN, 0.));
		pts[17] = this.newPoint(new RealPoint(b, 0., LEN, -LEN_2));
		pts[10] = this.newPoint(new RealPoint(b, -LEN_2, LEN, 0.));

		pts[6] = this.newPoint(new RealPoint(b, -LEN, 0., LEN_2));
		pts[11] = this.newPoint(new RealPoint(b, -LEN, LEN_2, 0.));
		pts[18] = this.newPoint(new RealPoint(b, -LEN, 0., -LEN_2));
		pts[12] = this.newPoint(new RealPoint(b, -LEN, -LEN_2, 0.));

		pts[7] = this.newPoint(new RealPoint(b, 0., -LEN, LEN_2));
		pts[13] = this.newPoint(new RealPoint(b, -LEN_2, -LEN, 0.));
		pts[19] = this.newPoint(new RealPoint(b, 0., -LEN, -LEN_2));
		pts[14] = this.newPoint(new RealPoint(b, LEN_2, -LEN, 0.));

		pts[20] = this.newPoint(new RealPoint(b, LEN_2, 0., -LEN));
		pts[21] = this.newPoint(new RealPoint(b, 0., LEN_2, -LEN));
		pts[22] = this.newPoint(new RealPoint(b, -LEN_2, 0., -LEN));
		pts[23] = this.newPoint(new RealPoint(b, 0., -LEN_2, -LEN));
		for (i = 0; i < 24; i++) {
				if (pts[i].x < 0 ||
				    pts[i].x >= this.width ||
				    pts[i].y < 0 ||
				    pts[i].y >= this.height)
					++out;
		}
		if (out == 24)
			return;

		/* Only draw those faces that are visible */
		d.x = this.v.x - b.x;
		d.y = this.v.y - b.y;
		d.z = this.v.z - b.z;
		/* back points looking through front
		   10  5  9
		 11    1    8
		  6  2   0  4
		 12    3   15
		   13  7 14
		  ----------
		   front points
		   10 17  9
		 11   21    8
		 18 22  20 16
		 12   23   15
		   13 19 14
		*/
		if (w || (cid[0] == null && cid[1] == null && cid[2] == null)) {
			if (d.z <= LEN)
				drawQuad(w, cid[2], bid, pts, 0, 1, 2, 3);
			else if (d.z >= -LEN)
				drawQuad(w, cid[2], bid, pts, 23, 22, 21, 20);
			if (d.y <= LEN)
				drawQuad(w, cid[1], bid, pts, 5, 9, 17, 10);
			else if (d.y >= -LEN)
				drawQuad(w, cid[1], bid, pts, 7, 14, 19, 13);
			if (d.x <= LEN)
				drawQuad(w, cid[0], bid, pts, 4, 8, 16, 15);
			else if (d.x >= -LEN)
				drawQuad(w, cid[0], bid, pts, 6, 11, 18, 12);
			if (d.x + d.y + d.z < 0) {
				drawHex(w, cid[3], bid, pts, 1, 0, 4, 8, 9, 5);
			} else {
				drawHex(w, cid[3], bid, pts, 22, 23, 19, 13, 12, 18);
			}
			if (d.x + d.y - d.z < 0) {
				drawHex(w, cid[1], bid, pts, 20, 21, 17, 9, 8, 16);
			} else {
				drawHex(w, cid[1], bid, pts, 3, 2, 6, 12, 13, 7);
			}
			if (d.x - d.y + d.z < 0) {
				drawHex(w, cid[0], bid, pts, 0, 3, 7, 14, 15, 4);
			} else {
				drawHex(w, cid[0], bid, pts, 21, 22, 18, 11, 10, 17);
			}
			if (-d.x + d.y + d.z < 0) {
				drawHex(w, cid[2], bid, pts, 2, 1, 5, 10, 11, 6);
			} else {
				drawHex(w, cid[2], bid, pts, 23, 20, 16, 15, 14, 19);
			}
		}
		if (d.x + d.y + d.z >= 0) {
			drawHexMouth(w, cid[3], bid, pts, 1, 0, 4, 8, 9, 5, dir == XPYPZP);
		} else {
			drawHexMouth(w, cid[3], bid, pts, 22, 23, 19, 13, 12, 18, dir == XNYNZN);
		}
		if (d.x + d.y - d.z >= 0) {
			drawHexMouth(w, cid[1], bid, pts, 20, 21, 17, 9, 8, 16, dir == XPYPZN);
		} else {
			drawHexMouth(w, cid[1], bid, pts, 3, 2, 6, 12, 13, 7, dir == XNYNZP);
		}
		if (d.x - d.y + d.z >= 0) {
			drawHexMouth(w, cid[0], bid, pts, 0, 3, 7, 14, 15, 4, dir == XPYNZP);
		} else {
			drawHexMouth(w, cid[0], bid, pts, 21, 22, 18, 11, 10, 17, dir == XNYPZN);
		}
		if (-d.x + d.y + d.z >= 0) {
			drawHexMouth(w, cid[2], bid, pts, 2, 1, 5, 10, 11, 6, dir == XNYPZP);
		} else {
			drawHexMouth(w, cid[2], bid, pts, 23, 20, 16, 15, 14, 19, dir == XPYNZN);
		}
		if (d.z > LEN)
			drawQuadMouth(w, cid[2], bid, pts, 0, 1, 2, 3, dir == ZP);
		else if (d.z < -LEN)
			drawQuadMouth(w, cid[2], bid, pts, 23, 22, 21, 20, dir == ZN);
		if (d.y > LEN)
			drawQuadMouth(w, cid[1], bid, pts, 5, 9, 17, 10, dir == YP);
		else if (d.y < -LEN)
			drawQuadMouth(w, cid[1], bid, pts, 7, 14, 19, 13, dir == YN);
		if (d.x > LEN)
			drawQuadMouth(w, cid[0], bid, pts, 4, 8, 16, 15, dir == XP);
		else if (d.x < -LEN)
			drawQuadMouth(w, cid[0], bid, pts, 6, 11, 18, 12, dir == XN);
	}

	/* This is for mapping mouse back to 3D position */
	/* Arash Partow <arashp AT hotmail.com>
	 * Homepage: http://www.partow.net
	 * Translated to C from Pascal
	 */
	static int orientation(Point a, Point b, Point p) {
		int orient;

		 /* Linear determinant of the 3 points */
		orient = (b.x - a.x) * (p.y - a.y) - (p.x - a.x) * (b.y - a.y);
		if (orient > 0)
			return 1; /* Orientation is to the right-hand side */
		else if (orient < 0)
			return -1; /* Orientation is to the left-hand side */
		else
			return 0; /* Orientation is neutral if result is 0 */
	}

	static boolean pointInTriangle(Point pt, Point a, Point b, Point c) {
		int or0 = orientation(a, b, pt);
		int or1 = orientation(b, c, pt);
		int or2 = orientation(c, a, pt);

		return (((or0 == or1) && (or1 == or2)) ||
			((or0 == 0) && (or1 == or2)) ||
			((or1 == 0) && (or0 == or2)) ||
			((or2 == 0) && (or0 == or1)));
	}

	static boolean pointInPolygon(Point pt, Point[] poly, int n) {
		int tri;
		Point a, b, c;

		a = poly[0];
		b = poly[1];
		for (tri = 2; tri < n; tri++) {
			c = poly[tri];
			if (pointInTriangle(pt, a, b, c))
				return true;
			b = c;
		}
		return false;
	}
}
